# flake8: noqa: E501

import logging
import os
import time
import warnings
from unittest.mock import patch
from pathlib import Path

import numpy as np
import pandas as pd
import pysnptools.util as pstutil
import scipy.stats as stats
from fastlmm.inference.fastlmm_predictor import (
    _kernel_fixup,
    _pheno_fixup,
    _snps_fixup,
    _SnpTrainTest,
)
from fastlmm.inference.lmm_cov import LMM as lmm_cov
from pysnptools.kernelreader import Identity as KernelIdentity
from pysnptools.kernelreader import KernelData, SnpKernel
from pysnptools.snpreader import Bed, Pheno, SnpData
from pysnptools.standardizer import DiagKtoN
from pysnptools.standardizer import Identity as SS_Identity
from pysnptools.standardizer import Standardizer, Unit
from pysnptools.util import create_directory_if_necessary
from pysnptools.util.intrangeset import IntRangeSet
from pysnptools.util.mapreduce1 import map_reduce


# !!!LATER add warning here (and elsewhere) K0 or K1.sid_count < test_snps.sid_count,
#  might be a covar mix up.(but only if a SnpKernel
def single_snp(
    test_snps,
    pheno,
    K0=None,
    K1=None,
    mixing=None,
    covar=None,
    covar_by_chrom=None,
    leave_out_one_chrom=True,
    output_file_name=None,
    h2=None,
    log_delta=None,
    show_snp_fract_var_exp=False,
    cache_file=None,
    GB_goal=None,
    interact_with_snp=None,
    force_full_rank=False,
    force_low_rank=False,
    G0=None,
    G1=None,
    runner=None,
    map_reduce_outer=True,
    pvalue_threshold=None,
    random_threshold=None,
    random_seed=0,
    xp=None,
    count_A1=None,
):
    """
    Function performing single SNP GWAS using cross validation over the chromosomes and REML. Will reorder and intersect IIDs as needed.
    (For backwards compatibility, you may use 'leave_out_one_chrom=False' to skip cross validation, but that is not recommended.)

    :param test_snps: SNPs to test. Can be any `SnpReader <http://fastlmm.github.io/PySnpTools/#snpreader-snpreader>`_.
           If you give a string, it should be the base name of a set of PLINK Bed-formatted files.
           (For backwards compatibility can also be dictionary with keys 'vals', 'iid', 'header')
    :type test_snps: a `SnpReader <http://fastlmm.github.io/PySnpTools/#snpreader-snpreader>`_ or a string

    :param pheno: A one or more phenotypes: Can be any `SnpReader <http://fastlmm.github.io/PySnpTools/#snpreader-snpreader>`_, for example,
           `Pheno <http://fastlmm.github.io/PySnpTools/#snpreader-pheno>`_ or `SnpData <http://fastlmm.github.io/PySnpTools/#snpreader-snpdata>`_.
           If you give a string, it should be the file name of a PLINK phenotype-formatted file.
           Any individual with missing all values will be removed.
           If more than one phenotype is given, then K1 (the 2nd kernel) cannot be given.
           (For backwards compatibility can also be dictionary with keys 'vals', 'iid', 'header')
    :type pheno: a `SnpReader <http://fastlmm.github.io/PySnpTools/#snpreader-snpreader>`_ or a string

    :param K0: similarity matrix or SNPs from which to create a similarity matrix. If not given, will use test_snps.
           Can be any `SnpReader <http://fastlmm.github.io/PySnpTools/#snpreader-snpreader>`_.
           If you give a string, it should be the base name of a set of PLINK Bed-formatted files.
           If leave_out_one_chrom is True, can be a dictionary from chromosome number to any `KernelReader <http://fastlmm.github.io/PySnpTools/#kernelreader-kernelreader>`_
           or the name a `KernelNpz <http://fastlmm.github.io/PySnpTools/#kernelreader-kernelnpz>`_-formated file.
           If leave_out_one_chrom is False, can be any `KernelReader <http://fastlmm.github.io/PySnpTools/#kernelreader-kernelreader>`_ or
           name of a `KernelNpz <http://fastlmm.github.io/PySnpTools/#kernelreader-kernelnpz>`_-formated file.
    :type K0: `SnpReader <http://fastlmm.github.io/PySnpTools/#snpreader-snpreader>`_ or a string
           or dictionary or `KernelReader <http://fastlmm.github.io/PySnpTools/#kernelreader-kernelreader>`_)

    :param K1: second similarity matrix or SNPs from which to create a second similarity matrix, optional. (Also, see 'mixing').
    :type K1: `SnpReader <http://fastlmm.github.io/PySnpTools/#snpreader-snpreader>`_ or a string
           or dictionary or `KernelReader <http://fastlmm.github.io/PySnpTools/#kernelreader-kernelreader>`_.

    :param mixing: Weight between 0.0 (inclusive, default) and 1.0 (inclusive) given to K1 relative to K0.
            If you give no mixing number and a K1 is given, the best weight will be learned.
    :type mixing: number

    :param covar: covariate information, optional: Can be any `SnpReader <http://fastlmm.github.io/PySnpTools/#snpreader-snpreader>`_, for example, `Pheno <http://fastlmm.github.io/PySnpTools/#snpreader-pheno>`_ or `SnpData <http://fastlmm.github.io/PySnpTools/#snpreader-snpdata>`_.
           If you give a string, it should be the file name of a PLINK phenotype-formatted file.
           Missing values are not supported.
           (For backwards compatibility can also be dictionary with keys 'vals', 'iid', 'header')
    :type covar: a `SnpReader <http://fastlmm.github.io/PySnpTools/#snpreader-snpreader>`_ or a string

    :param covar_by_chrom: dictionary from chromosome number to covariate information, optional:
           The covariate information
           can be any `SnpReader <http://fastlmm.github.io/PySnpTools/#snpreader-snpreader>`_.
           If given, leave_out_one_chrom must be True.
           Both covar and covar_by_chrom can be given.
           Missing values are not supported.
    :type covar_by_chrom: a `SnpReader <http://fastlmm.github.io/PySnpTools/#snpreader-snpreader>`_ or a string

    :param leave_out_one_chrom: Perform single SNP GWAS via cross validation over the chromosomes. Default to True.
           (Warning: setting False can cause proximal contamination.)
    :type leave_out_one_chrom: boolean

    :param output_file_name: Name of file to write results to, optional. If not given, no output file will be created. The output format is tab-delimited text.
    :type output_file_name: file name

    :param h2: A parameter to LMM learning, optional
            If not given will search for best value.
            If mixing is unspecified, then h2 must also be unspecified.
    :type h2: number

    :param log_delta: a re-parameterization of h2 provided for backwards compatibility. h2 is 1./(exp(log_delta)+1)
    :type log_delta: number

    :param show_snp_fract_var_exp: The output will always show 'EffectSize'. If show_snp_fract_var_exp is True, then the output will also show 'SnpFractVarExp'.
            EffectSize and SnpFractVarExp are different measures of fraction of variance explained.
            The first uses all variance of the phenotype in the denominator, whereas the second uses variance of the phenotype after
            removing the effects of population structure and family relatedness. Specifically,
            EffectSize is
            `SNPWeight^2 * "raw" test SNP variance / phenotype variance`, where:
            SNPWeight is the beta for the “standardized test SNP” (where "standardized" means made to have mean 0, stdev 1).
            And where "raw" means the original test SNP values of “0,1,2” (the count of allele 1 or 2 [it doesn’t matter]).
            And where the phenotype’s original values are used (so, they are not, for example, changed by regressing out the covariates).
    :type show_snp_fract_var_exp: boolean

    :param cache_file: Name of  file to read or write cached precomputation values to, optional.
                If not given, no cache file will be used.
                If given and file does not exist, will write precomputation values to file.
                If given and file does exist, will read precomputation values from file.
                The file contains the S and U from the decomposition of the training matrix and other values.
                It is in Python's np.savez (\*.npz) format.
                Calls using the same cache file should have the same inputs (pheno, K0, K1, covar) but test_snps can differ.
    :type cache_file: file name

    :param GB_goal: gigabytes of memory the run should use, optional. If not given, will read the test_snps in blocks the same size as the kernel,
        which is memory efficient with little overhead on computation time.
    :type GB_goal: number

    :param interact_with_snp: index of a covariate to perform an interaction test with.
            Allows for interaction testing (interact_with_snp x snp will be tested)
            default: None

    :param force_full_rank: Even if kernels are defined with fewer SNPs than IIDs, create an explicit iid_count x iid_count kernel. Cannot be True if force_low_rank is True.
    :type force_full_rank: Boolean

    :param force_low_rank: Even if kernels are defined with fewer IIDs than SNPs, create a low-rank iid_count x sid_count kernel. Cannot be True if force_full_rank is True.
    :type force_low_rank: Boolean

    :param G0: Same as K0. Provided for backwards compatibility. Cannot be given if K0 is given.
    :type G0: Same as K0.

    :param G1: Same as K1. Provided for backwards compatibility. Cannot be given if K1 is given.
    :type G1: Same as K1.

    :param runner: a `Runner <http://fastlmm.github.io/PySnpTools/#util-mapreduce1-runner-runner>`_, optional: Tells how to run locally, multi-processor, or on a cluster.
        If not given, the function is run locally.
    :type runner: `Runner <http://fastlmm.github.io/PySnpTools/#util-mapreduce1-runner-runner>`_

    :param pvalue_threshold: All output rows with p-values less than this threshold will be included. By default, all rows are included.
        This is used to exclude rows with large p-values.
    :type pvalue_threshold: number between 0 and 1

    :param random_threshold: All potential output rows are assigned a random value. All rows with a random value less than this threshold will be included.
        By default, all rows are included. This is used to create a random sample of output rows.
    :type random_threshold: number between 0 and 1

    :param random_seed: Seed used to assign a random values to rows. Used with random_threshold.
    :type random_threshold: integer

    :param map_reduce_outer: If true (default), divides work by chromosome. If false, divides test_snp work into chunks.
    :type map_reduce_outer: bool

    :param xp: The array module to use (optional), for example, 'numpy' (normal CPU-based module)
               or 'cupy' (GPU-based module). If not given, will try to read
               from the ARRAY_MODULE environment variable. If not given and ARRAY_MODULE is not set,
               will use numpy. If 'cupy' is requested, will
               try to 'import cupy'. If that import fails, will revert to numpy.
               If two kernels are given, will ignore this and use 'numpy'
    :type xp: string or Python module
    :rtype: Python module

    :param count_A1: If it needs to read SNP data from a BED-formatted file, tells if it should count the number of A1
         alleles (the PLINK standard) or the number of A2 alleles. False is the current default, but in the future the default will change to True.
    :type count_A1: bool

    :rtype: Pandas dataframe with one row per test SNP. Columns include "PValue"

    :Example:

    >>> import logging
    >>> from fastlmm.association import single_snp
    >>> from fastlmm.util import example_file # Download and return local file name
    >>> from pysnptools.snpreader import Bed
    >>> logging.basicConfig(level=logging.INFO)
    >>> pheno_fn = example_file("fastlmm/feature_selection/examples/toydata.phe")
    >>> test_snps = example_file("fastlmm/feature_selection/examples/toydata.5chrom.*","*.bed")
    >>> results_dataframe = single_snp(test_snps=test_snps, pheno=pheno_fn, count_A1=False)
    >>> print(results_dataframe.iloc[0].SNP,round(results_dataframe.iloc[0].PValue,7),len(results_dataframe))
    null_576 1e-07 10000

    For more examples, see:

    * `Main FaST-LMM notebook <https://github.com/fastlmm/FaST-LMM/blob/master/doc/ipynb/FaST-LMM.ipynb>`_
    * `Effect size, multiple phenotypes, and related new features <https://github.com/fastlmm/FaST-LMM/blob/master/doc/ipynb/fastlmm2021.ipynb>`_
    """
    #!!!LATER raise error if covar has NaN
    t0 = time.time()
    if force_full_rank and force_low_rank:
        raise Exception("Can't force both full rank and low rank")

    if K1 or G1:  # If 2nd kernel given, use numpy
        xp = "numpy"

    if output_file_name is not None:
        os.makedirs(Path(output_file_name).parent, exist_ok=True)

    xp = pstutil.array_module(xp)
    with patch.dict("os.environ", {"ARRAY_MODULE": xp.__name__}) as _:
        assert test_snps is not None, "test_snps must be given as input"
        test_snps = _snps_fixup(test_snps, count_A1=count_A1)
        pheno = _pheno_fixup(pheno, count_A1=count_A1).read()
        good_values_per_iid = (pheno.val == pheno.val).sum(axis=1)
        assert not np.any(
            (good_values_per_iid > 0) * (good_values_per_iid < pheno.sid_count)
        ), "With multiple phenotypes, an individual's values must either be all missing or have no missing."
        pheno = pheno[
            good_values_per_iid > 0, :
        ]  # drop individuals with no good pheno values.
        covar = _pheno_fixup(covar, iid_if_none=pheno.iid, count_A1=count_A1)

        if not leave_out_one_chrom:
            assert (
                covar_by_chrom is None
            ), "When 'leave_out_one_chrom' is False, 'covar_by_chrom' must be None"  # !!!LATER document covar_by_chrom
            K0 = _kernel_fixup(
                K0 or G0 or test_snps,
                iid_if_none=test_snps.iid,
                standardizer=Unit(),
                count_A1=count_A1,
            )
            K1 = _kernel_fixup(
                K1 or G1,
                iid_if_none=test_snps.iid,
                standardizer=Unit(),
                count_A1=count_A1,
            )
            K0, K1, test_snps, pheno, covar = pstutil.intersect_apply(
                [K0, K1, test_snps, pheno, covar]
            )
            logging.debug("# of iids now {0}".format(K0.iid_count))
            K0, K1, block_size = _set_block_size(
                K0, K1, mixing, GB_goal, force_full_rank, force_low_rank
            )

            frame = _internal_single(
                K0=K0,
                test_snps=test_snps,
                pheno=pheno,
                covar=covar,
                K1=K1,
                mixing=mixing,
                h2=h2,
                log_delta=log_delta,
                show_snp_fract_var_exp=show_snp_fract_var_exp,
                cache_file=cache_file,
                force_full_rank=force_full_rank,
                force_low_rank=force_low_rank,
                output_file_name=output_file_name,
                block_size=block_size,
                interact_with_snp=interact_with_snp,
                runner=runner,
                pvalue_threshold=pvalue_threshold,
                random_threshold=random_threshold,
                random_seed=random_seed,
                xp=xp,
            )
            if pvalue_threshold is None and random_threshold is None:
                sid_index_range = IntRangeSet(frame["sid_index"])
                assert sid_index_range == (
                    0,
                    test_snps.sid_count,
                ), "Some SNP rows are missing from the output"
        else:
            if map_reduce_outer:
                runner_outer = runner
                runner_inner = None
            else:
                runner_outer = None
                runner_inner = runner

            chrom_list = list(
                set(test_snps.pos[:, 0])
            )  # find the set of all chroms mentioned in test_snps, the main testing data
            assert not np.isnan(chrom_list).any(), "chrom list should not contain NaN"
            input_files = [test_snps, pheno, covar] + (
                [] if covar_by_chrom is None else list(covar_by_chrom.values())
            )
            for Ki in [K0, G0, K1, G1]:
                if isinstance(Ki, dict):
                    input_files += list(Ki.values())
                else:
                    input_files.append(Ki)

            def nested_closure(chrom):
                logging.info(f"working on chrom {chrom}")
                xp = pstutil.array_module()
                test_snps_chrom = test_snps[:, test_snps.pos[:, 0] == chrom]
                covar_chrom = _create_covar_chrom(covar, covar_by_chrom, chrom)
                cache_file_chrom = (
                    None if cache_file is None else f"{cache_file}.{chrom}.npz"
                )

                K0_chrom = _K_per_chrom(K0 or G0 or test_snps, chrom, test_snps.iid)
                K1_chrom = _K_per_chrom(K1 or G1, chrom, test_snps.iid)

                (
                    K0_chrom,
                    K1_chrom,
                    test_snps_chrom,
                    pheno_chrom,
                    covar_chrom,
                ) = pstutil.intersect_apply(
                    [K0_chrom, K1_chrom, test_snps_chrom, pheno, covar_chrom]
                )
                logging.debug("# of iids now {0}".format(K0_chrom.iid_count))
                K0_chrom, K1_chrom, block_size = _set_block_size(
                    K0_chrom, K1_chrom, mixing, GB_goal, force_full_rank, force_low_rank
                )

                distributable = _internal_single(
                    K0=K0_chrom,
                    test_snps=test_snps_chrom,
                    pheno=pheno_chrom,
                    covar=covar_chrom,
                    K1=K1_chrom,
                    mixing=mixing,
                    h2=h2,
                    log_delta=log_delta,
                    show_snp_fract_var_exp=show_snp_fract_var_exp,
                    cache_file=cache_file_chrom,
                    force_full_rank=force_full_rank,
                    force_low_rank=force_low_rank,
                    output_file_name=None,
                    block_size=block_size,
                    interact_with_snp=interact_with_snp,
                    runner=runner_inner,
                    pvalue_threshold=pvalue_threshold,
                    random_threshold=random_threshold,
                    random_seed=random_seed,
                    xp=xp,
                )
                return distributable

            def reducer_closure(frame_sequence):
                frame = pd.concat(frame_sequence)
                frame.sort_values(by="PValue", inplace=True)
                frame.index = np.arange(len(frame))
                if output_file_name is not None:
                    frame.to_csv(output_file_name, sep="\t", index=False)
                logging.info("PhenotypeName\t{0}".format(pheno.sid[0]))
                logging.info("SampleSize\t{0}".format(test_snps.iid_count))
                logging.info("SNPCount\t{0}".format(test_snps.sid_count))
                logging.info("Runtime\t{0}".format(time.time() - t0))

                return frame

            frame = map_reduce(
                chrom_list,
                mapper=nested_closure,
                reducer=reducer_closure,
                input_files=input_files,
                output_files=[output_file_name],
                name="single_snp (leave_out_one_chrom), out='{0}'".format(
                    output_file_name
                ),
                runner=runner_outer,
            )

    return frame


overhead_gig = 0.127
factor = 8.5  # found via trial and error


def _GB_goal_from_block_size(block_size, iid_count, kernel_gig):
    left_bytes = block_size * (iid_count * 8.0 * factor)
    left_gig = left_bytes / 1024.0**3
    GB_goal = left_gig + overhead_gig + kernel_gig
    return GB_goal


def _block_size_from_GB_goal(GB_goal, iid_count, min_count):
    kernel_bytes = iid_count * min_count * 8
    kernel_gig = kernel_bytes / (1024.0**3)

    if GB_goal is None:
        GB_goal = _GB_goal_from_block_size(min_count, iid_count, kernel_gig)
        logging.info("Setting GB_goal to {0} GB".format(GB_goal))
        return min_count

    left_gig = GB_goal - overhead_gig - kernel_gig
    if left_gig <= 0:
        warnings.warn(
            "The full kernel and related operations will likely not fit in the goal_memory"
        )
    left_bytes = left_gig * 1024.0**3
    snps_at_once = left_bytes / (iid_count * 8.0 * factor)
    block_size = int(snps_at_once)

    if block_size < min_count:
        block_size = min_count
        GB_goal = _GB_goal_from_block_size(block_size, iid_count, kernel_gig)
        warnings.warn(
            "Can't meet goal_memory without loading too few snps at once. Resetting GB_goal to {0} GB".format(
                GB_goal
            )
        )

    return block_size


def single_snp_leave_out_one_chrom(*args, **kwargs):
    """
    .. deprecated:: 0.2.22
       Use :meth:`single_snp` instead.

    """
    warnings.warn(
        "'single_snp_leave_out_one_chrom' is deprecated. Use 'single_snp(...) instead.",
        DeprecationWarning,
    )
    return single_snp(*args, **kwargs)


def _K_per_chrom(K, chrom, iid, count_A1=None):
    if K is None:
        return KernelIdentity(iid)
    else:
        if isinstance(K, dict):
            return _kernel_fixup(
                K[chrom], iid_if_none=iid, standardizer=Unit(), count_A1=count_A1
            )
        else:
            K_all = _kernel_fixup(
                K, iid_if_none=iid, standardizer=Unit(), count_A1=count_A1
            )
            if not isinstance(K_all, SnpKernel):
                raise Exception(
                    "Don't know how to make '{0}' work per chrom".format(K_all)
                )
            return SnpKernel(
                K_all.snpreader[:, K_all.pos[:, 0] != chrom], K_all.standardizer
            )


#!!!move to own file?
class _Mixer(object):
    def __init__(self, do_g, kernel_trained0, kernel_trained1, mixing):
        self.do_g = do_g
        self.kernel_trained0 = kernel_trained0
        self.kernel_trained1 = kernel_trained1
        self.mixing = mixing
        self.snp_trained0 = None
        self.snp_trained1 = None

    def k_mix(self, K0, K1):  #!!!later add special case code for mixing==1 and 0
        K0_b = K0.read().standardize(self.kernel_trained0)
        K1_b = K1.read().standardize(self.kernel_trained1)
        # similar code elsewhere
        K = np.empty(K0_b.val.shape)
        _mix_from_Ks(K, K0_b.val, K1_b.val, self.mixing)
        K = KernelData(val=K, iid0=K0_b.iid0, iid1=K0_b.iid1)
        return K

    # g_mix doesn't care about block_size because if we are low-rank the number SNPs is small.
    def g_mix(self, K0, K1):
        mixing = self.mixing

        if mixing == 1 or isinstance(K0, KernelIdentity):
            assert K1.standardizer is self.snp_trained1, "real assert"
            G_train = (
                K1.train.read()
                .standardize(self.snp_trained1)
                .standardize(self.kernel_trained1)
            )  #!!!later this a good place to read?
            G_test = (
                K1.test.read()
                .standardize(self.snp_trained1)
                .standardize(self.kernel_trained1)
            )  #!!!later this a good place to read?
            K = _SnpTrainTest(
                train=G_train, test=G_test, standardizer=SS_Identity(), block_size=None
            )
            return K

        if mixing == 0 or isinstance(K1, KernelIdentity):
            assert K0.standardizer is self.snp_trained0, "real assert"
            G_train = (
                K0.train.read()
                .standardize(self.snp_trained0)
                .standardize(self.kernel_trained0)
            )  #!!!later this a good place to read?
            G_test = (
                K0.test.read()
                .standardize(self.snp_trained0)
                .standardize(self.kernel_trained0)
            )  #!!!later this a good place to read?
            K = _SnpTrainTest(
                train=G_train, test=G_test, standardizer=SS_Identity(), block_size=None
            )
            return K

        #!!!later why are we processing the training data again????
        assert K0.standardizer is self.snp_trained0, "real assert"
        assert isinstance(K0, _SnpTrainTest), "Expect K0 to be a _SnpTrainTest"
        assert K1.standardizer is self.snp_trained1, "real assert"
        G0_train = (
            K0.train.read()
            .standardize(self.snp_trained0)
            .standardize(self.kernel_trained0)
        )  #!!!later this a good place to read?
        G1_train = (
            K1.train.read()
            .standardize(self.snp_trained1)
            .standardize(self.kernel_trained1)
        )  #!!!later this a good place to read?
        G0_test = (
            K0.test.read()
            .standardize(self.snp_trained0)
            .standardize(self.kernel_trained0)
        )  #!!!later this a good place to read?
        G1_test = (
            K1.test.read()
            .standardize(self.snp_trained1)
            .standardize(self.kernel_trained1)
        )  #!!!later this a good place to read?
        G_train = np.empty((K0.iid0_count, K0.train.sid_count + K1.train.sid_count))
        G_test = np.empty((K0.iid1_count, K0.train.sid_count + K1.train.sid_count))
        _mix_from_Gs(G_train, G0_train.val, G1_train.val, self.mixing)
        _mix_from_Gs(G_test, G0_test.val, G1_test.val, self.mixing)
        G_train = SnpData(
            iid=K0.iid0,
            sid=np.concatenate((K0.train.sid, K1.train.sid), axis=0),
            val=G_train,
            name="{0}&{1}".format(G0_train, G1_train),
            pos=np.concatenate((K0.train.pos, K1.train.pos), axis=0),
        )
        G_test = SnpData(
            iid=K0.iid1,
            sid=np.concatenate((K0.train.sid, K1.train.sid), axis=0),
            val=G_test,
            name="{0}&{1}".format(G0_test, G1_test),
            pos=np.concatenate((K0.train.pos, K1.train.pos), axis=0),
        )
        K = _SnpTrainTest(
            train=G_train, test=G_test, standardizer=SS_Identity(), block_size=None
        )
        return K

    def to_np(self):
        dict = {
            "do_g": self.do_g,
            #'G0_trained.stats':self.G0_trained.stats if self.G0_trained is not None else [],
            #'G1_trained.stats':self.G1_trained.stats if self.G1_trained is not None else [],
            "factor0": self.kernel_trained0,
            "factor1": self.kernel_trained1,
            "mixing": self.mixing,
        }
        return dict

    @staticmethod
    def from_np(data):
        mixer = _Mixer(
            do_g=bool(data["do_g"]),
            kernel_trained0=float(data["factor0"]),
            kernel_trained1=float(data["factor1"]),
            mixing=data["mixing"],
        )
        return mixer

    @staticmethod
    def combine_the_best_way(
        K0,
        K1,
        covar,
        y,
        mixing,
        h2,
        force_full_rank=False,
        force_low_rank=False,
        snp_standardizer=None,
        kernel_standardizer=None,
        block_size=None,
        xp=np,
    ):
        from pysnptools.kernelstandardizer import Identity as KS_Identity

        assert K0.iid0 is K0.iid1, "Expect K0 to be square"
        assert K1.iid0 is K1.iid1, "Expect K1 to be square"
        assert K0 is not None
        assert K1 is not None
        assert np.array_equal(
            K0.iid, K1.iid
        ), "Expect K0 and K1 to having matching iids"
        assert kernel_standardizer is not None, "expect values for kernel_standardizer"

        mixer = _Mixer(False, KS_Identity(), KS_Identity(), mixing)

        sid_count_0 = _Mixer.sid_counter(K0, force_full_rank, force_low_rank)
        sid_count_1 = _Mixer.sid_counter(K1, force_full_rank, force_low_rank)

        #################################
        # Both Identity (or not given)
        #################################
        if sid_count_0 + sid_count_1 == 0:
            h2 = h2 or 0
            mixer.mixing = mixer.mixing or 0
            K = (
                K0.read()
            )  # would be nice to use LinearRegression or low-rank with 0 snps

        #################################
        #
        #################################
        elif sid_count_0 + sid_count_1 < K0.iid_count or force_low_rank:
            mixer.do_g = True
            #!!!there is no need for block_size here because we want G0 in full. But if starting with SNPs and not low-rank then batches are needed and the two standardizers must be remembered for use later

            if sid_count_0 > 0:
                (
                    K0,
                    mixer.snp_trained0,
                    mixer.kernel_trained0,
                ) = K0._read_with_standardizing(
                    to_kerneldata=not mixer.do_g,
                    kernel_standardizer=kernel_standardizer,
                    return_trained=True,
                )
            if sid_count_1 > 0:
                (
                    K1,
                    mixer.snp_trained1,
                    mixer.kernel_trained1,
                ) = K1._read_with_standardizing(
                    to_kerneldata=not mixer.do_g,
                    kernel_standardizer=kernel_standardizer,
                    return_trained=True,
                )

            if sid_count_1 == 0:
                mixer.mixing = mixer.mixing or 0
                K = K0
            elif sid_count_0 == 0:
                mixer.mixing = mixer.mixing or 1
                K = K1
            else:
                if mixer.do_g:
                    G = xp.empty((K0.iid_count, K0.sid_count + K1.sid_count))
                    if mixer.mixing is None:
                        assert (
                            xp is np
                        ), "Currently, single_snp does not support cupy and searching for best mix of two kernels."
                        mixer.mixing, h2 = _find_mixing_from_Gs(
                            G, covar, K0.snpreader.val, K1.snpreader.val, h2, y, xp
                        )

                    if mixer.mixing == 0:
                        K = K0
                    elif mixer.mixing == 1:
                        K = K1
                    else:
                        _mix_from_Gs(
                            G, K0.snpreader.val, K1.snpreader.val, mixer.mixing
                        )
                        G = SnpData(
                            iid=K0.iid,
                            sid=["K0_{0}".format(i) for i in range(K0.sid_count)]
                            + [
                                "K1_{0}".format(i) for i in range(K1.sid_count)
                            ],  # rename the sids so that they can't collide.
                            val=G,
                            name="{0}&{1}".format(K0.snpreader, K1.snpreader),
                            pos=np.concatenate((K0.pos, K1.pos), axis=0),
                            xp=xp,
                        )
                        K = SnpKernel(G, SS_Identity(), block_size=block_size)
        else:
            mixer.do_g = False
            if (
                sid_count_0 > 0
            ):  #!!!but what if we have SNP data but still need to remember the standardizer?
                (
                    K0,
                    mixer.snp_trained0,
                    mixer.kernel_trained0,
                ) = K0._read_with_standardizing(
                    to_kerneldata=True, return_trained=True
                )  #!!!pass in a new argument, the kernel_standardizer(???)

            if sid_count_1 > 0:
                (
                    K1,
                    mixer.snp_trained1,
                    mixer.kernel_trained1,
                ) = K1._read_with_standardizing(to_kerneldata=True, return_trained=True)

            if sid_count_1 == 0:
                mixer.mixing = mixer.mixing or 0
                K = K0
            elif sid_count_0 == 0:
                mixer.mixing = mixer.mixing or 1
                K = K1
            else:
                K = xp.empty(K0.val.shape)
                if mixer.mixing is None:
                    assert (
                        xp is np
                    ), "Currently, single_snp does not support cupy and searching for best mix of two kernels."
                    mixer.mixing, h2 = _find_mixing_from_Ks(
                        K, covar, K0.val, K1.val, h2, y, xp
                    )
                _mix_from_Ks(K, K0.val, K1.val, mixer.mixing)
                assert (
                    K.shape[0] == K.shape[1]
                    and abs(xp.diag(K).sum() - K.shape[0]) < 1e-7
                ), "Expect mixed K to be standardized"
                K = KernelData(val=K, iid=K0.iid, xp=xp)

        return K, h2, mixer

    @staticmethod
    def sid_counter(K, force_full_rank, force_low_rank):
        if isinstance(K, KernelIdentity):
            return 0
        if force_full_rank:
            return np.inf
        if isinstance(K, SnpKernel):
            return K.sid_count
        else:
            return np.inf


def _set_block_size(K0, K1, mixing, GB_goal, force_full_rank, force_low_rank):
    min_count = _internal_determine_block_size(
        K0, K1, mixing, force_full_rank, force_low_rank
    )
    iid_count = K0.iid_count if K0 is not None else K1.iid_count
    block_size = _block_size_from_GB_goal(GB_goal, iid_count, min_count)
    # logging.info("Dividing SNPs by {0}".format(-(test_snps.sid_count//-block_size)))

    try:
        K0.block_size = block_size
    except:
        pass  # ignore

    try:
        K1.block_size = block_size
    except:
        pass  # ignore

    return K0, K1, block_size


def _internal_determine_block_size(K0, K1, mixing, force_full_rank, force_low_rank):
    assert not (force_full_rank and force_low_rank), "real assert"

    if isinstance(K0, SnpKernel) and K0.snpreader.sid_count == 0:
        K0 = KernelIdentity(K0.iid)
    if isinstance(K1, SnpKernel) and K1.snpreader.sid_count == 0:
        K1 = KernelIdentity(K1.iid)

    ##########################
    # A special case: both kernels are the Identity so just return the first one
    ##########################
    if isinstance(K0, KernelIdentity) and isinstance(K1, KernelIdentity):
        return K0.iid_count

    ##########################
    # Special cases: mixing says to use just one kernel or the other kernel is just identity, so just return one kernel
    ##########################
    if mixing == 0.0 or isinstance(K1, KernelIdentity):
        if (
            isinstance(K0, SnpKernel)
            and not force_full_rank
            and (force_low_rank or K0.snpreader.sid_count < K0.iid_count)
        ):
            return K0.snpreader.sid_count
        else:
            return K0.iid_count

    if mixing == 1.0 or isinstance(K0, KernelIdentity):
        if (
            isinstance(K1, SnpKernel)
            and not force_full_rank
            and (force_low_rank or K1.snpreader.sid_count < K1.iid_count)
        ):
            return K1.snpreader.sid_count
        else:
            return K1.iid_count

    ##########################
    # A special case: Treat the kernels as collections of snps (i.e. low-rank)
    ##########################
    if (
        isinstance(K0, SnpKernel)
        and isinstance(K1, SnpKernel)
        and not force_full_rank
        and (
            force_low_rank
            or K0.snpreader.sid_count + K1.snpreader.sid_count < K0.iid_count
        )
    ):
        return K0.snpreader.sid_count + K1.snpreader.sid_count

    ##########################
    # The most general case, treat the new kernels as kernels (i.e.. full rank)
    ##########################
    return K0.iid_count


def _create_dataframe(pvalue_count, show_snp_fract_var_exp):
    columns = [
        "sid_index",
        "SNP",
        "Chr",
        "GenDist",
        "ChrPos",
        "PValue",
        "SnpWeight",
        "SnpWeightSE",
        "EffectSize",
        "Mixing",
        "Nullh2",
    ]
    if show_snp_fract_var_exp:
        columns.insert(9, "SnpFractVarExpl")

    dataframe = pd.DataFrame(
        index=np.arange(pvalue_count),
        columns=columns,
    )
    #!!Is this the only way to set types in a dataframe?
    dataframe["sid_index"] = dataframe["sid_index"].astype(float)
    dataframe["Chr"] = dataframe["Chr"].astype(float)
    dataframe["GenDist"] = dataframe["GenDist"].astype(float)
    dataframe["ChrPos"] = dataframe["ChrPos"].astype(float)
    dataframe["PValue"] = dataframe["PValue"].astype(float)
    dataframe["SnpWeight"] = dataframe["SnpWeight"].astype(float)
    dataframe["SnpWeightSE"] = dataframe["SnpWeightSE"].astype(float)
    dataframe["EffectSize"] = dataframe["EffectSize"].astype(float)
    if show_snp_fract_var_exp:
        dataframe["SnpFractVarExpl"] = dataframe["SnpFractVarExpl"].astype(float)
    dataframe["Mixing"] = dataframe["Mixing"].astype(float)
    dataframe["Nullh2"] = dataframe["Nullh2"].astype(float)

    return dataframe


def _internal_single(
    K0,
    test_snps,
    pheno,
    covar,
    K1,
    mixing,
    h2,
    log_delta,
    show_snp_fract_var_exp,
    cache_file,
    force_full_rank,
    force_low_rank,
    output_file_name,
    block_size,
    interact_with_snp,
    runner,
    pvalue_threshold,
    random_threshold,
    random_seed,
    xp,
):
    assert K0 is not None, "real assert"
    assert K1 is not None, "real assert"
    assert block_size is not None, "real assert"
    assert mixing is None or 0.0 <= mixing <= 1.0
    if force_full_rank and force_low_rank:
        raise Exception("Can't force both full rank and low rank")

    assert (
        h2 is None or log_delta is None
    ), "if h2 is specified, log_delta may not be specified"
    if log_delta is not None:
        h2 = 1.0 / (xp.exp(log_delta) + 1)

    if h2 is not None and not isinstance(h2, np.ndarray):
        h2 = np.repeat(h2, pheno.shape[1])

    covar_val = xp.asarray(covar.read(view_ok=True, order="A").val)
    covar_val = xp.c_[
        covar_val, xp.ones((test_snps.iid_count, 1))
    ]  # view_ok because np.c_ will allocation new memory

    if interact_with_snp is not None:
        logging.info("interaction with %i" % interact_with_snp)
        assert (
            0 <= interact_with_snp < covar_val.shape[1] - 1
        ), "interact_with_snp is out of range"
        interact = covar_val[:, interact_with_snp].copy()
        interact -= interact.mean()
        interact /= interact.std()
    else:
        interact = None

    lmm, h2, mixing = _find_h2_s_u(
        mixing,
        h2,
        pheno,
        covar_val,
        xp,
        force_full_rank,
        force_low_rank,
        K0,
        K1,
        cache_file,
        runner,
    )
    assert lmm.Y.shape == pheno.shape, "expect pheno and lmm.Y to have the same shape"

    frame = _snp_tester(
        test_snps,
        interact,
        pheno,
        lmm,
        block_size,
        output_file_name,
        runner,
        h2,
        mixing,
        pvalue_threshold,
        random_threshold,
        random_seed,
        show_snp_fract_var_exp,
    )

    return frame


cache_version = 3


def save_cache(lmm_0, h2_0, mixing_0, cache_file, cache_file_extra, xp):
    pstutil.create_directory_if_necessary(cache_file)
    assert (
        lmm_0.U is not None and lmm_0.S is not None
    ), "Expect S and U have been computed"
    xp.savez(
        cache_file,
        version=cache_version,
        S=lmm_0.S,
        U=lmm_0.U,
        UY=lmm_0.UY,
        UUY=lmm_0.UUY if lmm_0.UUY is not None else [False],
        h2_0=h2_0,
        mixing_0=mixing_0,
    )
    if os.path.exists(cache_file_extra):
        os.unlink(cache_file_extra)


def load_cache(covar_val, y_0, cache_file, xp):
    lmm_0 = lmm_cov(X=covar_val, Y=y_0, G=None, K=None, xp=xp)
    with xp.load(cache_file) as data:  #!! similar code in epistasis
        version = data["version"]
        assert version == cache_version, f"Expect cache version {cache_version}"
        lmm_0.S = data["S"]
        lmm_0.U = data["U"]
        lmm_0.UY = data["UY"]
        lmm_0.UUY = None if data["UUY"].dtype != "float64" else data["UUY"]
        h2_0 = data["h2_0"]
        mixing_0 = data["mixing_0"]

    assert (
        len(h2_0) == 1
        and len(mixing_0.shape) == 0
        and y_0.shape[1] == 1
        and lmm_0.UY.shape[1] == 1
    ), "Expect cached h2_0, mixing_0, y_0, and lmm_0.UY to have dimension one"
    assert (
        lmm_0.UUY is None or lmm_0.UUY.shape[1] == 1
    ), "Expect lmm_0.UUY to have dimension 1"

    return lmm_0, h2_0, mixing_0


def save_cache_extra(lmm_multi, multi_h2, multi_mixing, cache_file_extra, xp):
    pstutil.create_directory_if_necessary(cache_file_extra)
    assert (
        lmm_multi.UY is not None and lmm_multi.S is not None
    ), "Expect UY have been computed"
    xp.savez(
        cache_file_extra,
        version=cache_version,
        UY=lmm_multi.UY,
        UUY=lmm_multi.UUY if lmm_multi.UUY is not None else [False],
        h2=multi_h2,
        mixing=multi_mixing,
    )


def load_cache_extra(lmm_multi, multi_y, cache_file_extra, xp):
    with xp.load(cache_file_extra) as data:  #!! similar code in epistasis
        version = data["version"]
        assert version == cache_version, f"Expect cache version {cache_version}"
        lmm_multi.UY = data["UY"]
        lmm_multi.UUY = None if data["UUY"].dtype != "float64" else data["UUY"]
        h2_multi = data["h2"]
        mixing_multi = data["mixing"]
    lmm_multi.Y = multi_y
    assert (
        len(h2_multi) == multi_y.shape[1]
        and lmm_multi.UY.shape[1] == multi_y.shape[1]
        and mixing_multi.shape[0] == multi_y.shape[1]
    ), "Expect cached h2, mixing, UY to agree with multi_y"
    assert (
        lmm_multi.UUY is None or lmm_multi.UUY.shape[1] == multi_y.shape[1]
    ), "Expect cached UUY to agree with multi_y"
    return lmm_multi, h2_multi, mixing_multi


def _find_h2_s_u(
    mixing,
    h2,
    multi_pheno,
    covar_val,
    xp,
    force_full_rank,
    force_low_rank,
    K0,
    K1,
    cache_file,
    runner,
):
    assert multi_pheno.sid_count >= 1, "Expect at least one phenotype"
    assert (
        isinstance(K1, KernelIdentity) or multi_pheno.sid_count == 1
    ), "When a 2nd kernel is given, only one phenotype is allowed."

    # view_ok because this code already did a fresh read to look for any missing values
    multi_y = xp.asarray(multi_pheno.read(view_ok=True, order="A").val)

    y_0 = multi_y[:, 0:1]
    cache_file_extra = f"{cache_file}.extra.npz" if cache_file is not None else None

    logging.info("Finding SU and then h2 for first phenotype")
    if cache_file is not None and os.path.exists(cache_file):
        lmm_0, h2_0, mixing_0 = load_cache(covar_val, y_0, cache_file, xp)
    else:
        lmm_0, h2_0, mixing_0 = _find_h2_s_u_for_one_pheno(
            K0,
            K1,
            covar_val,
            True,
            None,
            y_0,
            mixing,
            h2,
            force_full_rank,
            force_low_rank,
            None,
            None,
            None,
            xp,
        )

        if cache_file is not None:
            save_cache(lmm_0, h2_0, mixing_0, cache_file, cache_file_extra, xp)

    if multi_pheno.sid_count == 1:
        return lmm_0, h2_0, mixing_0
    elif cache_file_extra is not None and os.path.exists(cache_file_extra):
        return load_cache_extra(lmm_0, multi_y, cache_file_extra, xp)
    else:
        lmm_multi, multi_h2, multi_mixing = compute_extra(
            lmm_0,
            h2,
            h2_0,
            mixing_0,
            multi_y,
            multi_pheno,
            cache_file_extra,
            force_full_rank,
            force_low_rank,
            runner,
            xp,
        )
        if cache_file_extra is not None:
            save_cache_extra(lmm_multi, multi_h2, multi_mixing, cache_file_extra, xp)
        return lmm_multi, multi_h2, multi_mixing


def compute_extra(
    lmm_0,
    h2,
    h2_0,
    mixing_0,
    multi_y,
    multi_pheno,
    cache_file_extra,
    force_full_rank,
    force_low_rank,
    runner,
    xp,
):
    uy_list = [lmm_0.UY[:, 0]]
    if lmm_0.UUY is not None:
        uuy_list = [lmm_0.UUY[:, 0]]
    h2_list = [h2_0]
    mixing_list = [mixing_0]

    def mapper(pheno_index):
        logging.info(f"working on pheno_index {pheno_index} of {multi_pheno.sid_count}")
        return _find_h2_s_u_for_one_pheno(
            K0=None,
            K1=None,
            covar_val=lmm_0.X,
            regressX=lmm_0.regressX,
            linreg=lmm_0.linreg,
            y=multi_y[:, pheno_index : pheno_index + 1],
            mixing=mixing_0,
            h2=h2,
            force_full_rank=force_full_rank,
            force_low_rank=force_low_rank,
            K=lmm_0.K,
            S=lmm_0.S,
            U=lmm_0.U,
            xp=xp,
        )

    result_list = map_reduce(
        range(1, multi_pheno.col_count), mapper=mapper, runner=runner
    )
    # could do this with runner
    for lmm_p, h2_p, mixing_p in result_list:
        uy_list.append(lmm_p.UY[:, 0])
        if lmm_p.UUY is not None:
            uuy_list.append(lmm_p.UUY[:, 0])
        h2_list.append(h2_p)
        mixing_list.append(mixing_p)

    lmm_0.Y = multi_y
    lmm_0.UY = np.r_[uy_list].T
    if lmm_0.UUY is not None:
        lmm_0.UUY = np.r_[uuy_list].T
        assert (
            lmm_0.UUY.shape == multi_pheno.shape
        ), "expect pheno and lmm.UUY to have the same shape"
    multi_h2 = np.r_[h2_list]
    multi_mixing = np.r_[mixing_list]

    return lmm_0, multi_h2, multi_mixing


def _find_h2_s_u_for_one_pheno(
    K0,
    K1,
    covar_val,
    regressX,
    linreg,
    y,
    mixing,
    h2,
    force_full_rank,
    force_low_rank,
    K,
    S,
    U,
    xp,
):
    if S is None:
        K, h2, mixer = _Mixer.combine_the_best_way(
            K0,
            K1,
            covar_val,
            y,
            mixing,
            h2,
            force_full_rank=force_full_rank,
            force_low_rank=force_low_rank,
            kernel_standardizer=DiagKtoN(),
            xp=xp,
        )
        mixing = mixer.mixing

        if mixer.do_g:
            G = xp.asarray(K.snpreader.val)
            lmm = lmm_cov(
                X=covar_val,
                regressX=regressX,
                linreg=linreg,
                Y=y,
                K=None,
                G=G,
                inplace=True,
                S=None,
                U=None,
                xp=xp,
            )
        else:
            # print(covar_val.sum(),y.sum(),K.val.sum(),covar_val[0],y[0],K.val[0,0])
            lmm = lmm_cov(
                X=covar_val,
                regressX=regressX,
                linreg=linreg,
                Y=y,
                K=K.val,
                G=None,
                inplace=True,
                S=None,
                U=None,
                xp=xp,
            )
    else:
        lmm = lmm_cov(
            X=covar_val,
            regressX=regressX,
            linreg=linreg,
            Y=y,
            K=K,
            G=None,
            inplace=True,
            S=S,
            U=U,
            xp=xp,
        )

    if h2 is None:
        logging.info("Starting findH2")
        result = lmm.findH2()
        if not isinstance(result, list):
            result = [result]
        h2 = np.array([item["h2"] for item in result])

    logging.info("h2={0}".format(h2))

    return lmm, h2, mixing


def _snp_tester(
    test_snps,
    interact,
    pheno,
    lmm,
    block_size,
    output_file_name,
    runner,
    h2,
    mixing,
    pvalue_threshold,
    random_threshold,
    random_seed,
    show_snp_fract_var_exp,
):
    work_count = -(
        test_snps.sid_count // -block_size
    )  # Find the work count based on batch size (rounding up)
    pvalue_count = test_snps.sid_count * pheno.sid_count

    Sd, denom, h2 = lmm.get_Sd_etc(
        Sd=None, denom=None, h2=h2, logdelta=None, delta=None, scale=1, weightW=None
    )

    # We define three closures, that is, functions define inside function so that the inner function has access to the local variables of the outer function.
    def debatch_closure(work_index):
        return test_snps.sid_count * work_index // work_count

    def mapper_closure(work_index):
        xp = pstutil.array_module()
        if work_count > 1:
            logging.info(
                f"single_snp: Working on snp block {work_index} of {work_count}"
            )

        do_work_time = time.time()
        start = debatch_closure(work_index)
        end = debatch_closure(work_index + 1)

        raw_snps = test_snps[:, start:end]
        snps_read = raw_snps.read()
        if xp is np:
            snps_read.standardize()
            val = xp.asarray(snps_read.val)
        else:
            val = xp.asarray(snps_read.val)
            statsx = xp.empty(
                [val.shape[1], 2],
                dtype=val.dtype,
                order="F" if val.flags["F_CONTIGUOUS"] else "C",
            )
            Standardizer._standardize_unit_python(
                val, apply_in_place=True, use_stats=False, stats=statsx
            )

        if interact is not None:
            variables_to_test = val * interact[:, xp.newaxis]
        else:
            variables_to_test = val

        res = lmm.nLLeval(
            h2=h2,
            dof=None,
            scale=1.0,
            penalty=0.0,
            snps=variables_to_test,
            Sd=Sd,
            denom=denom,
        )

        assert test_snps.iid_count == lmm.U.shape[0]
        assert (
            res["beta"].size == (end - start) * pheno.sid_count
        ), "Expect multi_beta to be (end-start)x phenos"

        df = _multi_compute_stats(
            res["beta"],
            res["variance_beta"],
            res["fraction_variance_explained_beta"] if show_snp_fract_var_exp else None,
            start,
            end,
            raw_snps,
            snps_read,
            pheno.sid,
            mixing,
            h2,
            lmm,
            pvalue_threshold=pvalue_threshold,
            random_threshold=random_threshold,
            random_seed=random_seed,
            pvalue_count=pvalue_count,
            xp=xp,
        )

        logging.info("time={0}".format(time.time() - do_work_time))
        return df

    def reducer_closure(result_sequence):
        if output_file_name is not None:
            create_directory_if_necessary(output_file_name)

        frame = pd.concat(result_sequence)
        frame.sort_values(by="PValue", inplace=True)
        frame.index = np.arange(len(frame))

        if output_file_name is not None:
            frame.to_csv(output_file_name, sep="\t", index=False)

        return frame

    frame = map_reduce(
        range(work_count),
        mapper=mapper_closure,
        reducer=reducer_closure,
        input_files=[test_snps],
        output_files=[output_file_name],
        name="single_snp(output_file={0})".format(output_file_name),
        runner=runner,
    )
    return frame


def _multi_compute_stats(
    multi_beta,
    multi_variance_beta,
    multi_fraction_variance_explained_beta_or_none,
    start,
    end,
    raw_snps,
    snps_read,
    pheno_sid,
    mixing,
    h2,
    lmm,
    pvalue_threshold,
    random_threshold,
    random_seed,
    pvalue_count,
    xp,
):
    assert len(multi_beta.reshape(-1)) == (end - start) * len(
        pheno_sid
    ), "Expect multi_beta to be (end-start)x phenos"
    assert (
        multi_variance_beta.shape == multi_beta.shape
    ), "expect beta and variance_beta to agree on shape"
    assert (
        multi_fraction_variance_explained_beta_or_none is None
        or multi_beta.shape == multi_fraction_variance_explained_beta_or_none.shape
    ), "expect beta, and fraction_variance_explained_beta to agree on shape"

    g_var = np.nanvar(raw_snps.read().val, axis=0)
    p_var = lmm.Y.var(axis=0)
    effect_size = multi_beta**2 * g_var[..., None] / p_var[None, ...]

    chi2stats = pstutil.asnumpy(multi_beta * multi_beta / multi_variance_beta)
    p_values = stats.f.sf(chi2stats, 1, lmm.U.shape[0] - (lmm.linreg.D + 1))
    p_values = p_values.T.reshape(-1)
    pvalue_keep_index = p_values <= (pvalue_threshold or 1.0)
    if random_threshold is not None:
        rng = np.random.RandomState((start, random_seed))
        random_values = rng.random(len(pvalue_keep_index))
        random_keep_index = random_values <= random_threshold
        keep_index = random_keep_index + pvalue_keep_index
    else:
        keep_index = pvalue_keep_index
    dataframe = _create_dataframe(
        keep_index.sum(),
        show_snp_fract_var_exp=multi_fraction_variance_explained_beta_or_none
        is not None,
    )

    if len(pheno_sid) > 1:
        dataframe["Pheno"] = np.repeat(pheno_sid, snps_read.sid_count)[keep_index]
        dataframe["PhenoCount"] = len(pheno_sid)
    if pvalue_threshold is not None:
        dataframe["PValueThreshold"] = pvalue_threshold
    if random_threshold is not None:
        dataframe["RandomValue"] = random_values[keep_index]
        dataframe["RandomThreshold"] = random_threshold
        dataframe["RandomSeed"] = random_seed
    if pvalue_threshold is not None or random_threshold is not None:
        dataframe["PValueCount"] = pvalue_count

    dataframe["sid_index"] = np.tile(np.arange(start, end), len(pheno_sid))[keep_index]
    dataframe["SNP"] = np.tile(snps_read.sid, len(pheno_sid))[keep_index]
    dataframe["Chr"] = np.tile(snps_read.pos[:, 0], len(pheno_sid))[keep_index]
    dataframe["GenDist"] = np.tile(snps_read.pos[:, 1], len(pheno_sid))[keep_index]
    dataframe["ChrPos"] = np.tile(snps_read.pos[:, 2], len(pheno_sid))[keep_index]

    dataframe["PValue"] = p_values[keep_index]
    dataframe["SnpWeight"] = pstutil.asnumpy(multi_beta.T.reshape(-1))[keep_index]
    dataframe["SnpWeightSE"] = pstutil.asnumpy(
        xp.sqrt(multi_variance_beta.T.reshape(-1))
    )[keep_index]
    dataframe["EffectSize"] = effect_size.T.reshape(-1)[keep_index]
    if multi_fraction_variance_explained_beta_or_none is not None:
        dataframe["SnpFractVarExpl"] = pstutil.asnumpy(
            xp.sqrt(multi_fraction_variance_explained_beta_or_none.T.reshape(-1))
        )[keep_index]
    dataframe["Mixing"] = np.repeat(mixing, snps_read.sid_count)[keep_index]
    dataframe["Nullh2"] = np.repeat(h2, snps_read.sid_count)[keep_index]

    return dataframe


def _create_covar_chrom(covar, covar_by_chrom, chrom, count_A1=None):
    if covar_by_chrom is not None:
        covar_by_chrom_chrom = covar_by_chrom[chrom]
        covar_by_chrom_chrom = _pheno_fixup(
            covar_by_chrom_chrom, iid_if_none=covar, count_A1=count_A1
        )
        covar_after, covar_by_chrom_chrom = pstutil.intersect_apply(
            [covar, covar_by_chrom_chrom]
        )
        ret = SnpData(
            iid=covar_after.iid,
            sid=np.r_[covar_after.sid, covar_by_chrom_chrom.sid],
            val=np.c_[
                covar_after.read(order="A", view_ok=True).val,
                covar_by_chrom_chrom.read(order="A", view_ok=True).val,
            ],
        )  # view_ok because np.c_ will allocate new memory.
        return ret
    else:
        return covar


def _find_mixing_from_Gs(G, covar, G0_standardized_val, G1_standardized_val, h2, y, xp):
    logging.info("starting _find_mixing_from_Gs")
    import fastlmm.util.mingrid as mingrid

    assert h2 is None, "if mixing is None, expect h2 to also be None"
    resmin = [None]

    def f(
        mixing,
        G0_standardized_val=G0_standardized_val,
        G1_standardized_val=G1_standardized_val,
        covar=covar,
        y=y,
        **kwargs,
    ):
        if not isinstance(mixing, (int, int, float, complex)):
            assert mixing.ndim == 1 and mixing.shape[0] == 1
            mixing = mixing[0]

        _mix_from_Gs(G, G0_standardized_val, G1_standardized_val, mixing)
        lmm = lmm_cov(X=covar, Y=y, G=G, K=None, inplace=True, xp=xp)
        result = lmm.findH2()
        if (resmin[0] is None) or (result["nLL"] < resmin[0]["nLL"]):
            resmin[0] = result
        logging.info(
            "mixing_from_Gs\t{0}\th2\t{1}\tnLL\t{2}".format(
                mixing, result["h2"], result["nLL"]
            )
        )
        # logging.info("reporter:counter:single_snp,find_mixing_from_Gs_count,1")
        assert not np.isnan(result["nLL"]), "nLL should be a number (not a NaN)"
        return result["nLL"]

    mixing, nLL = mingrid.minimize1D(
        f=f, nGrid=10, minval=0.0, maxval=1.0, verbose=False
    )

    if not isinstance(mixing, (int, int, float, complex)):
        assert mixing.ndim == 1 and mixing.shape[0] == 1
        mixing = mixing[0]

    h2 = resmin[0]["h2"]
    return mixing, h2


def _find_mixing_from_Ks(K, covar, K0_val, K1_val, h2, y, xp):
    logging.info("starting _find_mixing_from_Ks")
    import fastlmm.util.mingrid as mingrid

    assert h2 is None, "if mixing is None, expect h2 to also be None"
    resmin = [None]

    def f(mixing, K0_val=K0_val, K1_val=K1_val, covar=covar, y=y, **kwargs):
        if not isinstance(mixing, (int, int, float, complex)):
            assert mixing.ndim == 1 and mixing.shape[0] == 1
            mixing = mixing[0]

        _mix_from_Ks(K, K0_val, K1_val, mixing)
        lmm = lmm_cov(X=covar, Y=y, G=None, K=K, inplace=True, xp=xp)
        result = lmm.findH2()
        if (resmin[0] is None) or (result["nLL"] < resmin[0]["nLL"]):
            resmin[0] = result
        logging.debug(
            "mixing_from_Ks\t{0}\th2\t{1}\tnLL\t{2}".format(
                mixing, result["h2"], result["nLL"]
            )
        )
        # logging.info("reporter:counter:single_snp,find_mixing_from_Ks_count,1")
        assert not np.isnan(result["nLL"]), "nLL should be a number (not a NaN)"
        return result["nLL"]

    mixing, nLL = mingrid.minimize1D(
        f=f, nGrid=10, minval=0.0, maxval=1.0, verbose=False
    )

    if not isinstance(mixing, (int, int, float, complex)):
        assert mixing.ndim == 1 and mixing.shape[0] == 1
        mixing = mixing[0]
    h2 = resmin[0]["h2"]
    return mixing, h2


#!!!later move these to _Mixing
def _mix_from_Gs(G, G0_standardized_val, G1_standardized_val, mixing):
    # logging.info("concat G1, mixing {0}".format(mixing))
    G[:, 0 : G0_standardized_val.shape[1]] = G0_standardized_val
    G[:, 0 : G0_standardized_val.shape[1]] *= np.sqrt(1.0 - float(mixing))
    G[:, G0_standardized_val.shape[1] :] = G1_standardized_val
    G[:, G0_standardized_val.shape[1] :] *= np.sqrt(float(mixing))


def _mix_from_Ks(K, K0_val, K1_val, mixing):
    K[:, :] = K0_val * (1.0 - float(mixing)) + K1_val * float(mixing)


if __name__ == "__main__":
    if True:
        logging.basicConfig(level=logging.WARN)

        import numpy as np
        from pysnptools.snpreader import Pheno, SnpData
        from fastlmm.util import example_file  # Download and return local file name

        # For example, create multiple phenotype in memory
        ##############################
        pheno1 = Pheno(example_file("tests/datasets/synth/pheno_10_causals.txt")).read()
        pheno3 = SnpData(
            iid=pheno1.iid,
            sid=["original", "sqrt", "square"],
            val=np.c_[pheno1.val, pheno1.val**2, pheno1.val**3],
        )

        # import the algorithm
        import numpy as np
        from fastlmm.association import single_snp
        from fastlmm.util import example_file  # Download and return local file name

        # set up data
        ##############################
        from fastlmm.util import example_file  # Download and return local file name

        bed_fn = example_file("tests/datasets/synth/all.*", "*.bed")
        cov_fn = example_file("tests/datasets/synth/cov.txt")

        # run gwas with pheno3
        ###################################################################
        results_df = single_snp(bed_fn, pheno3, covar=cov_fn, count_A1=False)
        print(results_df)

    if False:
        logging.basicConfig(level=logging.WARN)

        if True:
            from unittest.mock import patch

            from pysnptools.snpreader import Bed, DistributedBed
            from pysnptools.util.filecache import LocalCache

            seed = 1
            iid_count = 1 * 1000  # number of individuals
            sid_count = 5 * 1000  # number of SNPs
            cache_top = r"c:\deldir"
            leave_out_one_chrom = False
            runner = LocalMultiProc(5, just_one_process=False)

            if True:
                test_snps = Bed(
                    r"C:\Users\carlk\OneDrive\Shares\gaw14\gaw\gaw14smoking\synthetic\gaw14.1.Rand100k.familyMult\gaw14.1.Rand100K.familyMult.bed"
                )
                print(test_snps.shape)
            else:
                chrom_count = 10
                piece_per_chrom_count = 5  # Number of pieces for each chromosome
                file_cache_top = LocalCache(cache_top)

                test_snps_cache = file_cache_top.join(
                    "testsnps_{0}_{1}_{2}_{3}".format(
                        seed, chrom_count, iid_count, sid_count
                    )
                )

                if not next(
                    test_snps_cache.walk(), None
                ):  # If no files in the test_snps folder, generate data (takes about 6 hours)
                    from pysnptools.snpreader import SnpGen
                    from pysnptools.util.mapreduce1.runner import LocalMultiProc

                    snpgen = SnpGen(
                        seed=seed,
                        iid_count=iid_count,
                        sid_count=sid_count,
                        chrom_count=chrom_count,
                        block_size=1000,
                    )  # Create an on-the-fly SNP generator
                    snp_gen_runner = None  # LocalMultiProc(5)
                    # Write random SNP data to a DistributedBed
                    test_snps = DistributedBed.write(
                        test_snps_cache,
                        snpgen,
                        piece_per_chrom_count=piece_per_chrom_count,
                        runner=snp_gen_runner,
                    )
                else:
                    test_snps = DistributedBed(test_snps_cache)

            # Generate random pheno and covar
            import numpy as np
            from pysnptools.snpreader import SnpData

            np.random.seed(seed)
            pheno = SnpData(
                iid=test_snps.iid,
                sid=["pheno"],
                val=np.random.randn(test_snps.iid_count, 1) * 3 + 2,
            )
            covar = SnpData(
                iid=test_snps.iid,
                sid=["covar1", "covar2"],
                val=np.random.randn(test_snps.iid_count, 2) * 2 - 3,
            )

            test_snps.shape
            # both directions SVD
            # of say 5000 and more test snps than SVD snps
            # then add features (h2 search, multi kernel, crossval etc)
            from fastlmm.association import single_snp
            from pysnptools.snpreader import Bed

            logging.getLogger().setLevel(logging.WARN)
            print(logging.getLogger().level)
            # logging.getLogger("cloudpickle").setLevel(logging.INFO)

            array_module_name_list = ["numpy"]  # ,'cupy']
            print(array_module_name_list)
            for every in [1000, 500, 100, 50, 25, 10, 5, 4, 3, 2]:  #
                for GB_goal in [2]:  # .25,.5,1,2,4]:
                    K0 = test_snps[:, ::every]
                    print(
                        f"{iid_count}\t{sid_count}\t{K0.sid_count}\t{every}\t{os.environ.get('MKL_NUM_THREADS',12)}\t{GB_goal}\t{leave_out_one_chrom}",
                        end="",
                    )
                    # test_snps_cache = file_cache_top.join('testsnps_{0}_{1}_{2}_{3}'.format(seed,chrom_count,iid_count,sid_count))
                    # test_snps = DistributedBed(test_snps_cache)
                    # K0 = test_snps[:,::every] # re-creating this as a way to close it after K0.sid_count
                    for array_module_name in array_module_name_list:  #
                        # with patch.dict('os.environ', { #is this the generator that pickle/dill/cloudpickle don't like?
                        #    'ARRAY_MODULE': array_module_name,
                        #    }
                        #                ) as patched_environ:
                        # There are multiple ways to limit the # of threads and they must be set before 'import np'
                        # See https://stackoverflow.com/questions/30791550/limit-number-of-threads-in-numpy
                        # Here we just spot check that one has been set as expected.
                        assert os.environ["MKL_NUM_THREADS"] == "12"
                        # test_snps.close()
                        start = time.time()
                        results_dataframe = single_snp(
                            K0=K0,
                            test_snps=test_snps,
                            pheno=pheno,
                            covar=covar,
                            leave_out_one_chrom=leave_out_one_chrom,
                            count_A1=False,
                            GB_goal=GB_goal,
                            runner=runner,
                        )
                        # print(results_dataframe.iloc[0].SNP,round(results_dataframe.iloc[0].PValue,7),len(results_dataframe))
                        print(f"\t{time.time()-start}", end="")
                    print()

    if False:
        import logging

        from fastlmm.association import single_snp
        from fastlmm.util import example_file  # Download and return local file name
        from pysnptools.snpreader import Bed

        logging.basicConfig(level=logging.INFO)
        pheno_fn = example_file("fastlmm/feature_selection/examples/toydata.phe")
        test_snps = example_file(
            "fastlmm/feature_selection/examples/toydata.5chrom.*", "*.bed"
        )
        results_dataframe = single_snp(
            test_snps=test_snps, pheno=pheno_fn, count_A1=False
        )
        print(
            results_dataframe.iloc[0].SNP,
            round(results_dataframe.iloc[0].PValue, 7),
            len(results_dataframe),
        )

    if False:
        logging.basicConfig(level=logging.INFO)
        import doctest
        from unittest.mock import patch

        with patch.dict("os.environ", {"ARRAY_MODULE": "cupy"}) as _:
            doctest.testmod()

    if False:
        # Create a 2nd kernel, based just on the interesting pairs
        # Specifically, create SNP-like data for each individual
        # where the value is product of each pair's values
        # and the "SNP" name is "snp1,snp2"
        import os

        os.chdir("D:\OneDrive\Projects\Science")

        ###################################
        # Build the similarity matrix with everything except chroms 4 and 5.
        # Test every pair [first_n of] chrom 4 UNION [first n of] chrom 5
        chromA_of_interest = 4
        chromB_of_interest = 5
        first_n = 50  # None for all the chrom

        # import the algorithm and reader
        import numpy as np
        from fastlmm.association import epistasis
        from pysnptools.snpreader import Bed

        # define file names
        bed_reader = Bed("datasets/synth/all.bed", count_A1=True)
        pheno_fn = "datasets/synth/pheno_10_causals.txt"
        cov_fn = "datasets/synth/cov.txt"

        isin_A = bed_reader.pos[:, 0] == chromA_of_interest
        isin_B = bed_reader.pos[:, 0] == chromB_of_interest
        G0 = bed_reader[:, np.invert(isin_A + isin_B)]

        sid_list_AB = np.concatenate(
            [bed_reader.sid[isin_A][:first_n], bed_reader.sid[isin_B][:first_n]]
        )

        print(
            "similarity matrix created from {0:,} SNPs in chroms {1}".format(
                G0.sid_count, set(G0.pos[:, 0])
            )
        )
        a_count = len(bed_reader.sid[isin_A][:first_n])
        b_count = len(bed_reader.sid[isin_B][:first_n])
        print(
            "# of approx pairs={0:,}".format(
                a_count**2 // 2 + b_count**2 // 2 + a_count * b_count
            )
        )
        if False:
            ###################################
            # run epistasis analysis
            results_df = epistasis(
                bed_reader,
                pheno_fn,
                G0=G0,
                covar=cov_fn,
                sid_list_0=sid_list_AB,
                sid_list_1=sid_list_AB,
            )

            # qq plot
            from fastlmm.util.stats import plotp

            plotp.qqplot(results_df["PValue"].values, xlim=[0, 5], ylim=[0, 5])

            # print head of results data frame
            pd.set_option("display.width", 1000)
            results_df.head(n=10)
            ###################################
            interesting_pairs = results_df[0:5][["SNP0", "SNP1"]].values.tolist()
        else:
            interesting_pairs = [
                ["snp2741_m0_.23m1_.05", "snp2610_m0_.21m1_.22"],
                ["snp492_m0_.38m1_.36", "snp2124_m0_.17m1_.49"],
                ["snp325_m0_.45m1_.22", "snp2743_m0_.03m1_.1"],
                ["snp1060_m0_m1_.06", "snp2124_m0_.17m1_.49"],
                ["snp574_m0_.15m1_.03", "snp2124_m0_.17m1_.49"],
            ]

        from pysnptools.snpreader import SnpData

        npinteresting_pairs = np.array(interesting_pairs)  # convert to numpy array

        left_data = (
            bed_reader[:, bed_reader.sid_to_index(npinteresting_pairs[:, 0])]
            .read()
            .standardize()
        )
        right_data = (
            bed_reader[:, bed_reader.sid_to_index(npinteresting_pairs[:, 1])]
            .read()
            .standardize()
        )

        pair_val = left_data.val * right_data.val
        pair_sids = np.core.defchararray.add(
            np.core.defchararray.add(left_data.sid, ","), right_data.sid
        )

        pair_data = SnpData(
            val=pair_val, iid=bed_reader.iid, sid=pair_sids, name="interesting pairs"
        )
        pair_data

        test_snps = (
            bed_reader  # Should be all pairs of interest? just chrom 4 and 5 snps?
        )
        ss_results = single_snp(
            test_snps=bed_reader[:, 0],
            pheno=pheno_fn,
            covar=cov_fn,
            K0=G0,
            K1=pair_data,
            force_low_rank=True,
            leave_out_one_chrom=False,
        )
        # ss_results
        mixing, nullh2 = ss_results.iloc[0][["Mixing", "Nullh2"]]

        assert np.array_equal(G0.iid, pair_data.iid), "Expect matching iids"
        from pysnptools.snpreader import Pheno

        covar = Pheno(cov_fn).read()
        assert np.array_equal(G0.iid, covar.iid), "Expect matching iids"
        assert not np.any(np.isnan(covar.val))
        covar = np.c_[covar.val, np.ones((test_snps.iid_count, 1))]
        y = Pheno(pheno_fn).read()
        assert np.array_equal(y.iid, G0.iid), "Expect matching iids"
        assert not np.any(np.isnan(y.val))
        y = y.val

        from pysnptools.kernelreader import SnpKernel

        G_kernel = SnpKernel(G0, standardizer=Unit())
        E_kernel = SnpKernel(pair_data, standardizer=Unit())
        G_kernel = G_kernel.read().standardize()  # defaults to DiagKtoN standardize
        E_kernel = E_kernel.read().standardize()  # defaults to DiagKtoN standardize

        from fastlmm.inference.lmm import LMM

        lmm1 = LMM()
        lmm1.setK(K0=G_kernel.val, K1=E_kernel.val, a2=0.5)
        lmm1.setX(covar)
        lmm1.sety(y[:, 0])
        res1 = lmm1.findA2()
        h2, a2, nLLcorr = res1["h2"], res1["a2"], res1["nLL"]
        h2corr = h2 * (1 - a2)
        e2 = h2 * a2
        h2corr_raw = h2

    if False:
        import matplotlib.pyplot as plt
        from fastlmm.association import single_snp
        from pysnptools.snpreader import Bed

        plt.ylim()
        plt.show()

        snp_reader = Bed(
            r"M:\Temp\carl/hashdown/3f3ec259205678626199fb0ab3b0803a/tests/datasets/synth/all.bed"
        )
        pheno_fn = r"M:\Temp\carl\hashdown\3f3ec259205678626199fb0ab3b0803a\tests\datasets\synth\pheno_10_causals.txt"

        test_snps = snp_reader[:, snp_reader.pos[:, 0] == 1]
        results_dfx = single_snp(test_snps, pheno_fn)
        print(results_dfx.head())
        results_dfx = single_snp(test_snps, pheno_fn)
        print(results_dfx.head())



import logging
import numpy as np
import unittest
import os.path
import doctest
import pandas as pd
import pysnptools.util as pstutil
import sys

from fastlmm.inference import LinearRegression
from pysnptools.util.mapreduce1.runner import Local, LocalMultiProc
from pysnptools.snpreader import Dat, Bed, Pheno, SnpData
from fastlmm.feature_selection.test import TestFeatureSelection
from pysnptools.standardizer import Unit, Standardizer
from pysnptools.standardizer import Identity as SS_Identity
from pysnptools.kernelstandardizer import Identity as KS_Identity
from pysnptools.kernelreader import Identity as KernelIdentity
from pysnptools.kernelreader import KernelData, SnpKernel


class TestLinRegTrain(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from pysnptools.util import create_directory_if_necessary

        create_directory_if_necessary(self.tempout_dir, isfile=False)
        self.pythonpath = os.path.abspath(
            os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "..")
        )

        self.snpreader_whole = Bed(
            self.pythonpath + "/tests/datasets/synth/all.bed", count_A1=False
        )
        self.covariate_whole = Pheno(self.pythonpath + "/tests/datasets/synth/cov.txt")
        self.pheno_whole = Pheno(
            self.pythonpath + "/tests/datasets/synth/pheno_10_causals.txt"
        )

    tempout_dir = "tempout/linear_regression"

    def file_name(self, testcase_name):
        temp_fn = os.path.join(self.tempout_dir, testcase_name + ".dat")
        if os.path.exists(temp_fn):
            os.remove(temp_fn)
        return temp_fn

    def test_lr_real(self):
        do_plot = False

        import pylab

        logging.info("TestLinRegTrain test_lr_real")

        train_idx = np.r_[10 : self.snpreader_whole.iid_count]  # iids 10 and on
        test_idx = np.r_[0:10]  # the first 10 iids

        # make covar just numbers 0,1,...
        covar = self.covariate_whole.read()
        covar.val = np.array([[float(num)] for num in range(covar.iid_count)])
        covariate_train = covar[train_idx, :].read()
        covariate_test = covar[test_idx, :].read()
        K0_test_test = KernelIdentity(covariate_test.iid)

        # make pheno  # pheno = 2*covar+100+normal(0,1)*10
        pheno = self.pheno_whole.read()
        np.random.seed(0)
        pheno.val = covar.val * 2.0 + 100 + np.random.normal(size=covar.val.shape) * 10

        pheno_train = pheno[train_idx, :].read()
        pheno_test = pheno[test_idx, :].read()

        if do_plot:
            # Plot training x and y, testing x and y
            pylab.plot(
                covariate_train.val,
                pheno_train.val,
                ".",
                covariate_test.val,
                pheno_test.val,
                ".",
            )
            pylab.suptitle("Plot training x and y, testing x and y")
            pylab.show()

        Xtrain = np.c_[covariate_train.val, np.ones((covariate_train.iid_count, 1))]
        Xtest = np.c_[covariate_test.val, np.ones((covariate_test.iid_count, 1))]
        lsqSol = np.linalg.lstsq(Xtrain, pheno_train.val[:, 0], rcond=-1)
        bs = lsqSol[0]  # weights
        r2 = lsqSol[1]  # squared residuals
        D = lsqSol[2]  # rank of design matrix
        N = pheno_train.iid_count
        REML = False
        if not REML:
            sigma2 = float(r2 / N)
            nLL = N * 0.5 * np.log(2 * np.pi * sigma2) + N * 0.5
        else:
            sigma2 = float(r2 / (N - D))
            nLL = N * 0.5 * np.log(2 * np.pi * sigma2) + 0.5 / sigma2 * r2
            nLL -= 0.5 * D * np.log(2 * np.pi * sigma2)
            # REML term

        predicted = Xtest.dot(bs)
        yerr = [np.sqrt(sigma2)] * len(predicted)
        if do_plot:
            pylab.plot(
                covariate_test.val,
                pheno_test.val,
                "g.",
                covariate_test.val,
                predicted,
                "r.",
            )
            pylab.xlim([-1, 10])
            pylab.errorbar(covariate_test.val, predicted, yerr, linestyle="None")
            pylab.suptitle("real linear regression: actual to prediction")
            pylab.show()

        # These should all give the same result
        first_name = None
        for name, K0_train, K0_whole_test in [("Identity Kernel", None, None)]:

            first_name = first_name or name
            # Learn model, save, load
            modelx = LinearRegression().fit(
                K0_train=K0_train, X=covariate_train, y=pheno_train
            )

            filename = self.tempout_dir + "/model_lr_real.flm.p"
            pstutil.create_directory_if_necessary(filename)
            model = modelx

            do_test_on_train = True
            if do_test_on_train:
                # Predict with model (test on train)
                predicted_pheno, covar = model.predict(
                    K0_whole_test=K0_train, X=covariate_train
                )  # test on train
                output_file = self.file_name("lr_reala_" + name)
                Dat.write(output_file, predicted_pheno)
                covar2 = SnpData(
                    iid=covar.row, sid=covar.col[:, 1], val=covar.val
                )  # kludge to write kernel to text format
                output_file = self.file_name("lr_reala.cov_" + name)
                Dat.write(output_file, covar2)

                yerr = np.sqrt(np.diag(covar.val))
                predicted = predicted_pheno.val
                if do_plot:
                    pylab.plot(
                        covariate_train.val,
                        pheno_train.val,
                        "g.",
                        covariate_train.val,
                        predicted,
                        "r.",
                    )
                    pylab.xlim([0, 50])
                    pylab.ylim([100, 200])
                    pylab.errorbar(
                        covariate_train.val, predicted, yerr, linestyle="None"
                    )
                    pylab.suptitle(
                        name
                        + ": test on train: train X to true target (green) and prediction (red)"
                    )
                    pylab.show()

                self.compare_files(predicted_pheno, "lr2a_" + first_name)
                self.compare_files(covar2, "lr2a.cov_" + first_name)

            # Predict with model (test on test)
            predicted_pheno, covar = model.predict(
                K0_whole_test=K0_whole_test, X=covariate_test
            )  # test on train
            output_file = self.file_name("lr_realb_" + name)
            Dat.write(output_file, predicted_pheno)
            covar2 = SnpData(
                iid=covar.row, sid=covar.col[:, 1], val=covar.val
            )  # kludge to write kernel to text format
            output_file = self.file_name("lr_realb.cov_" + name)
            Dat.write(output_file, covar2)

            yerr = np.sqrt(np.diag(covar.val))
            predicted = predicted_pheno.val
            if do_plot:
                pylab.plot(
                    covariate_test.val,
                    pheno_test.val,
                    "g.",
                    covariate_test.val,
                    predicted,
                    "r.",
                )
                pylab.xlim([-1, 10])
                pylab.errorbar(covariate_test.val, predicted, yerr, linestyle="None")
                pylab.suptitle(
                    name
                    + ": test on test: test X to true target (green) and prediction (red)"
                )
                pylab.show()
                ## Plot y and predicted y (test on train)
                # pylab.plot(pheno_test.val,predicted_pheno.val,".")
                # pylab.suptitle(name+": test on test: true target to prediction")
                # pylab.show()

            self.compare_files(predicted_pheno, "lr2b_" + first_name)
            self.compare_files(covar2, "lr2b.cov_" + first_name)

    def compare_files(self, answer, ref_base):
        reffile = TestFeatureSelection.reference_file(
            "fastlmm/" + ref_base + ".dat"
        )  # Uses same results folder as lmm_train
        reference = Dat(reffile).read()
        assert np.array_equal(
            answer.col, reference.col
        ), "sid differs. File '{0}'".format(reffile)
        assert np.array_equal(
            answer.row, reference.row
        ), "iid differs. File '{0}'".format(reffile)
        for iid_index in range(reference.row_count):
            for sid_index in range(reference.col_count):
                a_v = answer.val[iid_index, sid_index]
                r_v = reference.val[iid_index, sid_index]
                assert (
                    abs(a_v - r_v) < 1e-4
                ), "Value at {0},{1} differs too much from file '{2}'".format(
                    iid_index, sid_index, reffile
                )

    def test_doctest(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/..")
        result = doctest.testmod(sys.modules["fastlmm.inference.linear_regression"])
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__


def getTestSuite():
    suite1 = unittest.TestLoader().loadTestsFromTestCase(TestLinRegTrain)
    return unittest.TestSuite([suite1])


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)

    # this import is needed for the runner
    # from fastlmm.inference.tests.test_linear_regression import TestLinRegTrain
    suites = unittest.TestSuite([getTestSuite()])

    if True:  # Standard test run
        r = unittest.TextTestRunner(failfast=False)
        ret = r.run(suites)
        assert ret.wasSuccessful()
    else:  # Cluster test run
        from pysnptools.util.mapreduce1.distributabletest import DistributableTest

        runner = HPC(
            10,
            "RR1-N13-09-H44",
            r"\\msr-arrays\Scratch\msr-pool\Scratch_Storage4\Redmond",
            remote_python_parent=r"\\msr-arrays\Scratch\msr-pool\Scratch_Storage4\REDMOND\carlk\Source\carlk\july_7_14\tests\runs\2014-07-24_15_02_02_554725991686\pythonpath",
            update_remote_python_parent=True,
            priority="AboveNormal",
            mkl_num_threads=1,
        )
        runner = Local()
        # runner = LocalMultiProc(taskcount=20,mkl_num_threads=5)
        # runner = LocalInParts(1,2,mkl_num_threads=1) # For debugging the cluster runs
        # runner = Hadoop(100, mapmemory=8*1024, reducememory=8*1024, mkl_num_threads=1, queue="default")
        distributable_test = DistributableTest(suites, "temp_test")
        print(runner.run(distributable_test))

    logging.info("done with testing")



# std modules
from collections import defaultdict
import gzip
import bz2
import pickle
import time
import os
import gc
import logging
import io

# common modules
import numpy as np
import pandas as pd
import sys

# sklearn
from sklearn.linear_model import RidgeCV, Ridge
from sklearn.model_selection import KFold, LeaveOneOut, ShuffleSplit
from sklearn.metrics import mean_squared_error
from sklearn.feature_selection import f_regression
from sklearn import model_selection
from sklearn.decomposition import PCA

# project
import fastlmm.pyplink.plink as plink
import pysnptools.util as pstutil
import fastlmm.util.util as util
import fastlmm.util.preprocess as up
import fastlmm.inference as fastlmm
import fastlmm.inference.linear_regression as lin_reg
from fastlmm.pyplink.snpset import *

from pysnptools.snpreader import Bed


class PerformSelectionDistributable(object):  # implements IDistributable
    def __init__(
        self,
        feature_selection_strategy,
        k_values,
        delta_values,
        strategy,
        output_prefix=None,
        select_by_ll=False,
        penalty=0.0,
        create_pdf=True,
    ):
        self.feature_selection_strategy = feature_selection_strategy
        self.k_values = k_values
        self.delta_values = delta_values
        self.strategy = strategy
        self.output_prefix = output_prefix
        self.select_by_ll = select_by_ll
        self.penalty = penalty
        self.num_steps_k = len(self.k_values)
        self.num_steps_delta = len(self.delta_values)
        self.create_pdf = create_pdf

    def copyinputs(self, copier):
        copier.input(self.feature_selection_strategy)

    def copyoutputs(self, copier):
        if self.output_prefix is not None:
            if self.feature_selection_strategy.num_pcs > 0:
                copier.output(
                    self.feature_selection_strategy.pcs_filename(self.output_prefix)
                )
            copier.output(self.output_prefix + "_snp.csv")
            copier.output(self.output_prefix + "_report.txt")
            if (self.strategy == "lmm_full_cv") or (self.strategy == "insample_cv"):
                for label in ["mse", "ll"]:
                    out_fn = self.output_prefix + "_" + label + ".csv"
                    copier.output(out_fn)
                    if self.create_pdf:
                        xval_fn = self.output_prefix + "_xval_%s.pdf" % label
                        copier.output(xval_fn)
                # if (self.strategy=="lmm_full_cv") and self.feature_selection_strategy.interpolate_delta:
                #    plot_fn=self.output_prefix+"_parabola.pdf"
                #    copier.output(plot_fn)
            else:  # this code branch is not regression tested, because it can't be reached.
                plot_fn = self.output_prefix + "_plot.pdf"
                copier.output(plot_fn)

    # start of IDistributable interface--------------------------------------
    @property
    def work_count(self):
        return (
            self.feature_selection_strategy.num_folds + 1
        )  # the +1 is for the linreg on the entire dataset

    def work_sequence(self):
        for fold_idx in range(self.feature_selection_strategy.num_folds):
            yield lambda fold_idx=fold_idx: self.dowork(
                fold_idx
            )  # the 'fold_idx=fold_idx is need to get around a strangeness in Python
        yield lambda output_prefix=self.output_prefix: self.feature_selection_strategy.linreg_entire_dataset(
            output_prefix
        )

    def reduce(self, result_sequence):
        """

        TODO: finish docstring
        return

        """
        # self.feature_selection_strategy.run_once() #Don't need and save time

        ##########################################
        ## perform model selection
        ##########################################

        mse_cv = (
            []
        )  # np.zeros((self.feature_selection_strategy.num_folds, len(self.k_values), len(self.delta_values)))
        ll_cv = (
            []
        )  # np.zeros((self.feature_selection_strategy.num_folds, len(self.k_values), len(self.delta_values)))
        best_delta_for_k = (
            []
        )  # np.zeros((self.feature_selection_strategy.num_folds, len(self.k_values)))
        lingreg_results = None
        for result in result_sequence:
            if (
                len(result) == 2
            ):  # the lingreg_results look different than the regular results because they are length two
                if lingreg_results is not None:
                    raise Exception("assert")
                lingreg_results = result
                continue
            fold_idx, mse_cv1, ll_cv1, best_delta_for_k_1 = result
            mse_cv.append(mse_cv1)
            ll_cv.append(ll_cv1)
            best_delta_for_k.append(best_delta_for_k_1)
        if lingreg_results == None:
            raise Exception("assert")
        if (self.strategy == "insample_cv") or (self.strategy == "lmm_full_cv"):
            if (
                len(ll_cv) != len(mse_cv)
                or len(mse_cv) != self.feature_selection_strategy.num_folds
            ):
                raise Exception(
                    "These should be the same. len(ll_cv)={0}, len(mse_cv)={1}, self.feature_selection_strategy.num_folds={2}".format(
                        len(ll_cv),
                        len(mse_cv),
                        self.feature_selection_strategy.num_folds,
                    )
                )
        else:
            assert (
                len(best_delta_for_k)
                == len(mse_cv)
                == self.feature_selection_strategy.num_folds
            )

        # find best parameters
        if mse_cv is not None:
            (
                best_k_mse,
                best_delta_mse,
                best_mse,
                best_delta_mse_interp,
                best_mse_interp,
            ) = self.feature_selection_strategy.reduce_result(
                mse_cv,
                self.k_values,
                self.delta_values,
                self.strategy,
                self.output_prefix,
                best_delta_for_k,
                label="mse",
                create_pdf=self.create_pdf,
            )
        if ll_cv is not None:
            (
                best_k_ll,
                best_delta_ll,
                best_ll,
                best_delta_ll_interp,
                best_ll_interp,
            ) = self.feature_selection_strategy.reduce_result(
                ll_cv,
                self.k_values,
                self.delta_values,
                self.strategy,
                self.output_prefix,
                best_delta_for_k,
                label="ll",
                create_pdf=self.create_pdf,
            )

        if self.select_by_ll:
            best_k, best_delta, best_obj, best_delta_interp, best_obj_interp = (
                best_k_ll,
                best_delta_ll,
                best_ll,
                best_delta_ll_interp,
                best_ll_interp,
            )
        else:
            best_k, best_delta, best_obj, best_delta_interp, best_obj_interp = (
                best_k_mse,
                best_delta_mse,
                best_mse,
                best_delta_mse_interp,
                best_mse_interp,
            )

        # perform final scan on whole data set
        best_snps, sorted_pval = self.feature_selection_strategy.final_scan(
            best_k, lingreg_results
        )

        # write report file
        if self.output_prefix is not None:
            report = "k_grid: " + str([k for k in self.k_values]) + "\n"
            ln_delta_grid = np.array([np.log(x) for x in self.delta_values])
            report += (
                "ln_delta_grid: ["
                + ", ".join([f"{v:.1f}" for v in ln_delta_grid])
                + "]\n"
            )
            report += "best k=%i\nbest ln_delta=%.1e\nbest objective=%.2f" % (
                best_k,
                np.log(best_delta),
                best_obj,
            )
            if (
                self.feature_selection_strategy.interpolate_delta
                and best_delta_interp is not None
            ):
                report += "\nbest ln_delta_interp=%.1e\nbest objective_interp=%.2f" % (
                    np.log(best_delta_interp),
                    best_obj_interp,
                )

            report_fn = self.output_prefix + "_report.txt"
            pstutil.create_directory_if_necessary(report_fn)
            report_file = open(report_fn, "w")
            report_file.write(report)
            report_file.close()

            # write out SNPs to keep
            # df = pd.DataFrame({"snp_id": best_snps.index, "snp_rs": best_snps.values, "p_values": sorted_pval})
            df = pd.DataFrame({"snp_rs": best_snps.values})  # Change snp_rs to sid?
            df.to_csv(self.output_prefix + "_snp.csv", index=False, header=False)

        return best_k, best_delta, best_obj, best_snps

    @property
    def tempdirectory(self):
        if self.output_prefix is None:
            return ".work_directory"
        else:
            return self.output_prefix + ".work_directory"

    def __str__(self):
        if self.output_prefix == None:
            return self.__class__.__name__
        else:
            return "{1} {0}".format(self.__class__.__name__, self.output_prefix)

    # end of IDistributable interface---------------------------------------

    def __repr__(self):
        fp = io.StringIO()
        fp.write("{0}(\n".format(self.__class__.__name__))
        varlist = []
        for f in dir(self):
            if f.startswith("_"):  # remove items that start with '_'
                continue
            if (
                type(self.__class__.__dict__.get(f, None)) is property
            ):  # remove @properties
                continue
            if callable(getattr(self, f)):  # remove methods
                continue
            varlist.append(f)
        for var in varlist[:-1]:  # all but last
            fp.write("\t{0}={1},\n".format(var, getattr(self, var).__repr__()))
        var = varlist[-1]  # last
        fp.write("\t{0}={1})\n".format(var, getattr(self, var).__repr__()))
        result = fp.getvalue()
        fp.close()
        return result

    def dowork(self, fold_idx):
        self.feature_selection_strategy.run_once()
        for i_k, k in enumerate(self.k_values):
            self.k_values[i_k] = min(
                self.k_values[i_k], self.feature_selection_strategy.snpreader.sid_count
            )
        max_k = max(
            [1]
            + [
                k
                for k in self.k_values
                if k != self.feature_selection_strategy.snpreader.sid_count
            ]
        )

        split_iterator = self.feature_selection_strategy.setup_linear_regression(
            max_k, start=fold_idx, stop=None
        )
        fold_data = next(split_iterator)

        tt0 = time.time()

        if self.strategy == "lmm_full_cv":
            mse_cv1 = np.zeros((len(self.k_values), len(self.delta_values)))
            ll_cv1 = np.zeros((len(self.k_values), len(self.delta_values)))
            best_delta_for_k_1 = None
        elif self.strategy == "insample_cv":
            mse_cv1 = np.zeros((len(self.k_values)))
            ll_cv1 = np.zeros((len(self.k_values)))
            best_delta_for_k_1 = np.zeros((len(self.k_values)))
        else:
            raise NotImplementedError("not implemented")

        logging.info("reporter:counter:PerformSelectionDistributable,foldcount,1")
        for k_idx, k in enumerate(self.k_values):
            logging.info("processing fold={0}, k={1}".format(fold_idx, k))
            logging.info(
                "reporter:status:processing fold={0}, k={1}".format(fold_idx, k)
            )
            logging.info("reporter:counter:PerformSelectionDistributable,k,1")

            model = fastlmm.getLMM()

            # compute kernel externally
            if (
                k == self.feature_selection_strategy.snpreader.sid_count
                or k >= self.feature_selection_strategy.num_snps_in_memory
            ):
                if k == self.feature_selection_strategy.snpreader.sid_count:
                    # use precomputed kernel
                    logging.info("using precomputed kernel on all snps")
                    K = self.feature_selection_strategy.K
                else:
                    # build kernel in blocks from snpreader (from file)
                    logging.info("building kernel in blocks")
                    top_k_feat_idx = fold_data["feat_idx"][0 : int(k)]
                    subset = self.feature_selection_strategy.snpreader[
                        :, top_k_feat_idx
                    ]
                    K = subset.kernel(
                        self.feature_selection_strategy.standardizer,
                        blocksize=self.feature_selection_strategy.blocksize,
                    )

                train_idx = fold_data["train_idx"]
                test_idx = fold_data["test_idx"]

                K_train_lhs = K[train_idx]
                K_train = K_train_lhs[:, train_idx]
                K_train_test = K_train_lhs[:, test_idx].T
                K_test_test = K[test_idx][:, test_idx]

                model.setK(K_train)
                model.setTestData(Xstar=fold_data["X_test"], K0star=K_train_test)

                # np.testing.assert_array_almost_equal(model.K, K_train, decimal=4)
                # np.testing.assert_array_almost_equal(model.Kstar, K_train_test, decimal=4)

            # use precomputed features as before
            else:
                logging.info("using cached data to build kernel")
                outer_G_train = fold_data["G_train"][:, 0:k]
                outer_G_test = fold_data["G_test"][:, 0:k]
                model.setG(outer_G_train.val)
                model.setTestData(Xstar=fold_data["X_test"], G0star=outer_G_test.val)
                K_test_test = None

            model.sety(fold_data["y_train"])
            model.setX(fold_data["X_train"])

            if self.strategy == "lmm_full_cv":
                for delta_idx, delta_act in enumerate(self.delta_values):
                    if k:
                        delta = delta_act * k
                    else:
                        delta = delta_act
                    REML = True  # TODO: Why is REML False?

                    # predict on test set
                    res = model.nLLeval(delta=delta, REML=REML, penalty=self.penalty)
                    out = model.predictMean(beta=res["beta"], delta=delta)
                    mse_cv1[k_idx, delta_idx] = mean_squared_error(
                        fold_data["y_test"], out
                    )
                    ll_cv1[k_idx, delta_idx] = model.nLLeval_test(
                        fold_data["y_test"],
                        res["beta"],
                        sigma2=res["sigma2"],
                        delta=delta,
                        Kstar_star=K_test_test,
                    )

            elif self.strategy == "insample_cv":
                best_res = None
                best_delta = None
                best_nLL = float("inf")
                REML = True

                # Note that for brent = True there will always be many unique delta values, as these deiate from the grid.
                # brent = False
                brent = True

                # evaluate negative log-likelihood for different values of alpha
                import fastlmm.util.mingrid as mingrid

                resmin = [None]

                def f(x):
                    if k:
                        delta_corr = x * k
                    else:
                        delta_corr = x
                    myres = model.nLLeval(
                        delta=delta_corr, REML=REML, penalty=self.penalty
                    )
                    if (resmin[0] is None) or (myres["nLL"] < resmin[0]["nLL"]):
                        resmin[0] = myres
                        resmin[0]["delta_corr"] = delta_corr
                        resmin[0]["delta"] = x
                    return myres["nLL"]

                res = mingrid.minimize1D(f, evalgrid=self.delta_values, brent=brent)

                if 0:  # old code without brent search
                    for delta_idx, delta_act in enumerate(self.delta_values):
                        delta = delta_act * k  # rescale delta for X val.
                        res = model.nLLeval(
                            delta=delta, REML=REML, penalty=self.penalty
                        )
                        # TODO: check if we need scale

                        if res["nLL"] < best_nLL:
                            best_res = res
                            best_delta_act = delta_act
                            best_delta = delta
                            best_nLL = res["nLL"]
                out = model.predictMean(
                    beta=resmin[0]["beta"], delta=resmin[0]["delta_corr"]
                )
                mse_cv1[k_idx] = mean_squared_error(fold_data["y_test"], out)
                ll_cv1[k_idx] = model.nLLeval_test(
                    fold_data["y_test"],
                    resmin[0]["beta"],
                    sigma2=resmin[0]["sigma2"],
                    delta=resmin[0]["delta_corr"],
                    Kstar_star=K_test_test,
                )
                best_delta_for_k_1[k_idx] = resmin[0]["delta"]

        logging.info("crossval time %.2f s" % (float(time.time() - tt0)))

        return fold_idx, mse_cv1, ll_cv1, best_delta_for_k_1


def build_kernel_blocked(
    snpreader, snp_idx=None, blocksize=10000, alt_snpreader=None, allowlowrank=False
):
    """build kernel by loading blocks of SNPs"""
    if alt_snpreader is None:
        alt_snpreader = snpreader

    if hasattr(alt_snpreader, "ind_used") and alt_snpreader.ind_used is not None:
        N = len(alt_snpreader.ind_used)
    else:
        N = len(alt_snpreader.original_iids)

    t0 = time.time()

    K = np.zeros([N, N])
    num_snps = alt_snpreader.snp_count

    if snp_idx is not None:
        snp_names = alt_snpreader.rs[snp_idx]
        current_size = len(snp_names)
        logging.info(
            "reading %i SNPs in blocks of %i and adding up kernels"
            % (len(snp_idx), blocksize)
        )
    else:
        current_size = num_snps
        logging.info(
            "constructing K from all %i SNPs (for %i individuals)" % (num_snps, N)
        )

    ct = 0
    ts = time.time()

    if (not allowlowrank) and alt_snpreader.snp_count < N:
        raise Exception("need to adjust code to handle low rank")

    for start in range(0, current_size, blocksize):
        ct += blocksize

        if snp_idx == None:
            tmp_set = PositionRange(start, blocksize)
        else:
            tmp_set = SnpAndSetName("someset", snp_names[start : start + blocksize])

        snps = alt_snpreader.read(tmp_set)["snps"]
        snps = up.standardize(snps)

        # logging.info("start = {0}".format(start))
        K += snps.dot(snps.T)

        if ct % blocksize == 0:
            logging.info("read %s SNPs in %.2f seconds" % (ct, time.time() - ts))

    # normalize kernel
    # K = K/np.sqrt(alt_snpreader.snp_count)

    # K = K + 1e-5*np.eye(N,N)
    t1 = time.time()
    logging.info("%.2f seconds elapsed" % (t1 - t0))

    return K



"""
file to set up python package, see http://docs.python.org/2/distutils/setupscript.html for details.
"""


import platform
import os
import sys
import shutil

from distutils.core import setup
from distutils.command.clean import clean as Clean

try:
    import numpy
except Exception:
    print("numpy needed for installation, please install numpy first")
    sys.exit()


def readme():
    with open('README.md') as f:
       return f.read()


class CleanCommand(Clean):
    description = "Remove build directories, and compiled files (including .pyc)"

    def run(self):
        Clean.run(self)
        if os.path.exists('build'):
            shutil.rmtree('build')


setup(
    name='fastlmm',
    version='0.1',
    description='Fast GWAS',
    long_description=readme(),
    keywords='gwas bioinformatics LMMs MLMs',
    url='',
    author='FaST-LMM Team',
    author_email='...',
    packages=[
        "fastlmm/association/tests",
        "fastlmm/association",
        "fastlmm/external/sklearn/externals",
        "fastlmm/external/sklearn/metrics",
        "fastlmm/external/sklearn",
        "fastlmm/external/util",
        "fastlmm/external",
        "fastlmm/feature_selection",
        "fastlmm/inference/bingpc",
        "fastlmm/inference",
        "fastlmm/pyplink/altset_list",
        "fastlmm/pyplink/snpreader",
        "fastlmm/pyplink/snpset",
        "fastlmm/pyplink",
        "fastlmm/util/stats",
        "fastlmm/util",
        "fastlmm"
    ],
    install_requires=['numpy', 'scipy', 'pandas', 'scikit-learn', 'matplotlib'],
    cmdclass = {'clean': CleanCommand},
    include_dirs = [numpy.get_include()],
  )




import fastlmm.util.preprocess as util


def TESTBEFOREUSING_score_testfilesFromDir(
    phenofile,
    base0,
    pedfilesalt,
    covarfile=None,
    outfile=None,
    ipheno=0,
    mindist=-1.0,
    idist=2,
    filetype="PED",
):
    """
    given a list of basefilenames that define alternative models and a basefilename of the null model
    test all alternative models
    --------------------------------------------------------------------------
    Input:
    phenofile   : filename of the phenotype file
    base0       : basefilename of the .ped and .map files containing the
                  null-model SNPs
    pedfilesalt : [Nalt] list of basefilenames of the Nalt .ped and .map files
                  containing the alternative-model SNPs
    covarfile   : filename of the covariates file (default None, in this case
                  only a bias is used)
    outfile     : filename of the output file (default None, in this case no
                  output is written to disk)
    ipheno      : 0-based index of the phenotype to be analyzed (default 0)
    mindist     : minimum distance for SNPs to be included in null model
                  (default -1.0: no excluson in this case)
    idist       : index in pos array that the exclusion is based on.
                  (1=genetic distance, 2=basepair distance)
    filetype    : plink filetype of the input (default 'PED')
                    'PED'   : PED file format
                    'BED'   : BED file format
    --------------------------------------------------------------------------
    Output dictionary:
    'pv'        : [Nalt] array P-values,
    'lik0'      : [Nalt] array containing the model parameters and negative
                  log-likelihoods of the null models,
    'lik1'      : [Nalt] array containing the model parameters and negative
                  log-likelihoods of the alternative models,
    'nexclude'  : [Nalt] array of numbers of SNPs excluded,
    'filenames' : [Nalt] array of basefilenames
    --------------------------------------------------------------------------
    """
    pheno = pstpheno.loadPhen(filename=phenofile, missing="-9", pheno=None)
    if covarfile is None:
        X = np.ones((pheno["vals"].shape[0], 1))
    else:
        covar = pstpheno.loadPhen(filename=covarfile, missing="-9", pheno=None)
        X = np.hstack((np.ones((pheno["vals"].shape[0], 1)), covar["vals"]))
    if filetype == "PED":
        SNPs0 = plink.readPED(
            basefilename=base0, delimiter=" ", missing="0", standardize=True, pheno=None
        )
    elif filetype == "BED":
        SNPs0 = plink.readBED(basefilename=base0)
        SNPs0["snps"] = util.standardize(SNPs0["snps"])

    y = pheno["vals"][:, ipheno]
    G0 = SNPs0["snps"] / np.sqrt(SNPs0["snps"].shape[1])

    # build the null model
    test2K = scoretest(Y=y[:, np.newaxis], X=X, K=None, G=G0)

    squaredform = np.zeros(len(pedfilesalt))
    expectationsqform = np.zeros(len(pedfilesalt))
    varsqform = np.zeros(len(pedfilesalt))
    squaredform2K = np.zeros(len(pedfilesalt))
    expectationsqform2K = np.zeros(len(pedfilesalt))
    varsqform2K = np.zeros(len(pedfilesalt))
    nexclude = np.zeros(len(pedfilesalt))
    include = np.zeros(len(pedfilesalt))
    Pv = np.zeros(len(pedfilesalt))
    Pv2K = np.zeros(len(pedfilesalt))

    for i, base1 in enumerate(pedfilesalt):  # iterate over all ped files
        SNPs1 = plink.readPED(
            basefilename=base1, delimiter=" ", missing="0", standardize=True, pheno=None
        )
        if mindist >= 0:
            i_exclude = excludeinds(
                SNPs0["pos"], SNPs1["pos"], mindist=mindist, idist=idist
            )
            nexclude[i] = i_exclude.sum()
        else:
            nexclude[i] = 0
        G1 = SNPs1["snps"] / np.sqrt(SNPs1["snps"].shape[1])

        if nexclude[i] > 0:
            test2Ke = scoretest(Y=y[:, np.newaxis], X=X, K=None, G=G0[:, ~i_exclude])
            squaredform2K[i], expectationsqform2K[i], varsqform2K[i] = test2Ke.score(
                G=G1
            )
        else:
            squaredform2K[i], expectationsqform2K[i], varsqform2K[i] = test2K.score(
                G=G1
            )
        squaredform[i], expectationsqform[i], varsqform[i] = scoreNoK(
            y, X=X, G=G1, sigma2=None
        )

        # perform moment matching
        Pv2K[i], dofchi22K, scalechi22K = pv_mom(
            squaredform2K[i], expectationsqform2K[i], varsqform2K[i]
        )
        Pv[i], dofchi2, scalechi2 = pv_mom(
            squaredform[i], expectationsqform[i], varsqform[i]
        )

    ret = {
        "filenames": np.array(pedfilesalt, dtype=str),
        "squaredform": squaredform,
        "expectationsqform": expectationsqform,
        "varsqform": varsqform,
        "P": Pv,
        "squaredform2K": squaredform2K,
        "expectationsqform2K": expectationsqform2K,
        "varsqform2K": varsqform2K,
        "nexclude": nexclude,
        "P2K": Pv2K,
    }
    if outfile is not None:
        # TODO
        print("implement me!")
        # header = np.array(['PV_5050','neg_log_lik_0','neg_loglik_alt','n_snps_excluded','filename_alt'])
        # data = np.concatenate(())
    return ret


def lrt_testfilesFromDir(
    phenofile,
    base0,
    pedfilesalt,
    covarfile=None,
    outfile=None,
    ipheno=0,
    mindist=-1.0,
    idist=2,
    filetype="PED",
):
    """
    given a list of basefilenames that define alternative models and a basefilename of the null model
    test all alternative models
    --------------------------------------------------------------------------
    Input:
    phenofile   : filename of the phenotype file
    base0       : basefilename of the .ped and .map files containing the
                  null-model SNPs
    pedfilesalt : [Nalt] list of basefilenames of the Nalt .ped and .map files
                  containing the alternative-model SNPs
    covarfile   : filename of the covariates file (default None, in this case
                  only a bias is used)
    outfile     : filename of the output file (default None, in this case no
                  output is written to disk)
    ipheno      : 0-based index of the phenotype to be analyzed (default 0)
    mindist     : minimum distance for SNPs to be included in null model
                  (default -1.0: no excluson in this case)
    idist       : index in pos array that the exclusion is based on.
                  (1=genetic distance, 2=basepair distance)
    filetype    : plink filetype of the input (default 'PED')
                    'PED'   : PED file format
                    'BED'   : BED file format
    --------------------------------------------------------------------------
    Output dictionary:
    'pv'        : [Nalt] array P-values,
    'lik0'      : [Nalt] array containing the model parameters and negative
                  log-likelihoods of the null models,
    'lik1'      : [Nalt] array containing the model parameters and negative
                  log-likelihoods of the alternative models,
    'nexclude'  : [Nalt] array of numbers of SNPs excluded,
    'filenames' : [Nalt] array of basefilenames
    --------------------------------------------------------------------------
    """
    pheno = pstpheno.loadPhen(filename=phenofile, missing="-9", pheno=None)
    if covarfile is None:
        X = np.ones((pheno["vals"].shape[0], 1))
    else:
        covar = pstpheno.loadPhen(filename=covarfile, missing="-9", pheno=None)
        X = np.hstack((np.ones((pheno["vals"].shape[0], 1)), covar["vals"]))
    if filetype == "PED":
        SNPs0 = plink.readPED(
            basefilename=base0, delimiter=" ", missing="0", standardize=True, pheno=None
        )
    elif filetype == "BED":
        SNPs0 = plink.readBED()
    y = pheno["vals"][:, ipheno]
    G0 = SNPs0["snps"] / np.sqrt(SNPs0["snps"].shape[1])

    # build the null model
    a0 = fastlmm.getLMM()
    a0.setG(G0)
    a0.setX(X)
    a0.sety(y)
    lik0_default = (
        a0.findH2()
    )  # The null model only has a single kernel and only needs to find h2
    lik0 = np.zeros(len(pedfilesalt), dtype="object")
    lik1 = np.zeros(len(pedfilesalt), dtype="object")
    lrt = np.zeros(len(pedfilesalt))
    pv = np.zeros(len(pedfilesalt))
    nexclude = np.zeros(len(pedfilesalt))
    for i, base1 in enuemrate(pedfilesalt):  # iterate over all ped files
        SNPs1 = plink.readPED(
            basefilename=base1, delimiter=" ", missing="0", standardize=True, pheno=None
        )
        i_exclude = excludeinds(
            SNPs0["pos"], SNPs1["pos"], mindist=mindist, idist=idist
        )
        nexclude[i] = i_exclude.sum()
        G1 = SNPs1["snps"] / np.sqrt(SNPs1["snps"].shape[1])
        if nexclude[i]:  # recompute the null likelihood
            G0_excluded = SNPs0["snps"][:, ~i_exclude] / np.sqrt(
                SNPs0["snps"][:, ~i_exclude].shape[1]
            )
            [pv[i], lik0[i], lik1[i]] = twokerneltest(
                G0=G0_excluded, G1=G1, y=y, covar=X, appendbias=False, lik0=None
            )
        else:  # use precomputed null likelihood
            [pv[i], lik0[i], lik1[i]] = twokerneltest(
                G0=G0, G1=G1, y=y, covar=X, appendbias=False, lik0=lik0_default
            )
    ret = {
        "pv": pv,
        "lik0": lik0,
        "lik1": lik1,
        "nexclude": nexclude,
        "filenames": np.array(pedfilesalt, dtype=str),
    }
    if outfile is not None:
        # TODO
        print("implement me!")
        # header = np.array(['PV_5050','neg_log_lik_0','neg_loglik_alt','n_snps_excluded','filename_alt'])
        # data = np.concatenate(())
    return ret



import sys
import numpy as np
import logging
import unittest
import os.path
import doctest
import pandas as pd
import sys

from fastlmm.association import single_snp
from fastlmm.association import single_snp_linreg
import pysnptools.util.pheno as pstpheno
from fastlmm.feature_selection.test import TestFeatureSelection
from pysnptools.util.mapreduce1.runner import Local, LocalMultiProc
from pysnptools.kernelreader import Identity as KernelIdentity
from pysnptools.standardizer import Unit
from pysnptools.snpreader import Bed, Pheno, SnpData
from pysnptools.kernelreader import SnpKernel


class TestSingleSnpLinReg(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        from pysnptools.util import create_directory_if_necessary

        create_directory_if_necessary(self.tempout_dir, isfile=False)
        self.pythonpath = os.path.abspath(
            os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "..")
        )
        self.bedbase = os.path.join(
            self.pythonpath, "tests/datasets/all_chr.maf0.001.N300"
        )
        self.phen_fn = os.path.join(
            self.pythonpath, "tests/datasets/phenSynthFrom22.23.N300.randcidorder.txt"
        )
        self.cov_fn = os.path.join(
            self.pythonpath, "tests/datasets/all_chr.maf0.001.covariates.N300.txt"
        )

    tempout_dir = "tempout/single_snp_linreg"

    def file_name(self, testcase_name):
        temp_fn = os.path.join(self.tempout_dir, testcase_name + ".txt")
        if os.path.exists(temp_fn):
            os.remove(temp_fn)
        return temp_fn

    def test_linreg(self):
        logging.info("TestSingleSnpLinReg test_linreg")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn

        output_file = self.file_name("linreg")

        frame1 = single_snp_linreg(
            test_snps=test_snps[:, :10],
            pheno=pheno,
            covar=covar,
            output_file_name=output_file,
            count_A1=False,
        )

        frame1 = frame1[["sid_index", "SNP", "Chr", "GenDist", "ChrPos", "PValue"]]
        self.compare_files(frame1, "linreg")

        frame2 = single_snp_linreg(
            test_snps=test_snps[:, :10],
            pheno=pheno,
            covar=covar,
            output_file_name=output_file,
            count_A1=False,
        )
        self.compare_files(frame2, "linreg")

    def compare_files(self, frame, ref_base):
        reffile = TestFeatureSelection.reference_file(
            "single_snp/" + ref_base + ".txt"
        )  # Results are in single_snp, not single_snp_lin_reg

        # sid_list,pvalue_list = frame['SNP'].values,frame['Pvalue'].values

        # sid_to_pvalue = {}
        # for index, sid in enumerate(sid_list):
        #    sid_to_pvalue[sid] = pvalue_list[index]

        reference = pd.read_csv(reffile, delimiter="\s", comment=None, engine="python")
        assert len(frame) == len(
            reference
        ), "# of pairs differs from file '{0}'".format(reffile)
        for _, row in reference.iterrows():
            sid = row.SNP
            pvalue = frame[frame["SNP"] == sid].iloc[0].PValue
            assert (
                abs(row.PValue - pvalue) < 1e-5
            ), "pair {0} differs too much from file '{1}'".format(sid, reffile)

    def test_doctest(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/..")
        result = doctest.testmod(sys.modules["fastlmm.association.single_snp_linreg"])
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__


def getTestSuite():

    suite1 = unittest.TestLoader().loadTestsFromTestCase(TestSingleSnpLinReg)
    return unittest.TestSuite([suite1])


if __name__ == "__main__":

    # this import is needed for the runner
    from fastlmm.association.tests.test_single_snp_linreg import TestSingleSnpLinReg

    suites = unittest.TestSuite([getTestSuite()])

    if True:  # Standard test run
        r = unittest.TextTestRunner(failfast=False)
        ret = r.run(suites)
        assert ret.wasSuccessful()
    else:  # Cluster test run
        logging.basicConfig(level=logging.INFO)

        from pysnptools.util.mapreduce1.distributabletest import DistributableTest

        runner = Local()
        # runner = LocalMultiProc(taskcount=20,mkl_num_threads=5)
        # runner = LocalInParts(1,2,mkl_num_threads=1) # For debugging the cluster runs
        # runner = Hadoop(100, mapmemory=8*1024, reducememory=8*1024, mkl_num_threads=1, queue="default")
        distributable_test = DistributableTest(suites, "temp_test")
        print(runner.run(distributable_test))

    logging.info("done with testing")



from fastlmm.association.FastLmmSet import *

# from pysnptools.util.mapreduce1.distributable import *
from pysnptools.util.mapreduce1.runner import *
import os
import sys
import time
import logging

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    logging.info("last modified: %s" % time.ctime(os.path.getmtime(__file__)))
    logging.info("-------------------------------------------------")

    # Parse the command line
    if len(sys.argv) == 2:
        sys.argv.append("Local()")
    if len(sys.argv) != 3:
        if len(sys.argv) != 5 and not "chrm" in sys.argv:
            logging.info(
                "Usage:python FastLmmSetLOOC.py <inputfile> [optional <runner>]"
            )
            sys.exit(1)
    file, infile, runner_string = sys.argv[0:3]
    if not os.path.exists(infile):
        raise Exception("Fatal Error: " + infile + " does not exist")

    # Run the job specified in infile
    from fastlmm.association.FastLmmSet import *

    exec("runner = " + runner_string)
    d = open(infile).read()
    exec(d)

    if "chrm" in sys.argv:
        i = sys.argv.index("chrm")
        mychrom = int(sys.argv[i + 1])
        chrmset = [mychrom]
    else:
        raise Exception("please specify 'chrm' in the input file")

    for chr in chrmset:
        print("doing leave-one-out chromosome")
        print("chrm: " + str(chr))
        tt0 = time.time()

        if bedfilealt.rfind("%i") > -1:
            bedfilealtnew = bedfilealt % chr
        else:
            bedfilealtnew = bedfilealt
        if outfile.rfind("%i") > -1:
            outfile = outfile % chr
        if filenull is not None and filenull.rfind("%i") > -1:
            filenullnew = filenull % chr
        else:
            filenullnew = filenull

        distributable = FastLmmSet(
            phenofile=phenofile,
            outfile=outfile,
            filenull=filenullnew,
            extractSim=extractSim,
            autoselect=False,
            bedfilealt=bedfilealtnew,
            altset_list=altset_list,
            covarfile=covarfile,
            mpheno=mpheno,
            mindist=mindist,
            idist=idist,
            nperm=nperm,  # for calibration of LRT test statistic
            test=test,
            nullfit=nullfit,
            qmax=qmax,
            datestamp=datestamp,
            nullModel=nullModel,
            altModel=altModel,
            genphen=None,  # for synthetic experiments
            permute=permute,  # to permute SNPs being tested: set to None on real data
            write_lrtperm=write_lrtperm,
            nullfitfile=nullfitfile,  # if cached calibration runs to re-use (no need on real data)
            sets=sets,
            calseed=calseed,  # for different random seeds on null model permutation calibration
            minsetsize=minsetsize,
            maxsetsize=maxsetsize,
            covarimp=covarimp,  # if want to impute the covariates
        )

        runner.run(distributable)
        tt1 = time.time()
        logging.info(
            "Final elapsed time for all processing is %.2f seconds" % (tt1 - tt0)
        )



exactUpdate = False
topKbyLinReg = 16
logdelta = np.log(16)

if topKbyLinReg is not None:
    outFile = "examples/ScanISP.exact%d.nSnps%d.logdelta%.2f.out.txt" % (
        exactUpdate,
        topKbyLinReg,
        logdelta,
    )
else:
    outFile = "examples/ScanISP.exact%d.all.logdelta%.2f.out.txt" % (
        exactUpdate,
        logdelta,
    )

obj = ScanISP(
    phenoFile="examples/toydataTrain.phe",
    # phenoTestFile = 'examples/toydataTest.phe',
    bedFile="examples/toydata",
    # fastlmmPath = 'C:/Users/t-baraki/Users/t-baraki/adaptivelmm/software/FaSTLMM.205.Win/Cpp_MKL',
    fastlmmPath=r"C:\Users\lippert\Projects\FaST-LMM_old\FastLmm\CPP\x64\MKL-Release",
    windowByPosition=1e1,
    logdelta=logdelta,
    topKbyLinReg=topKbyLinReg,
    exactUpdate=exactUpdate,
    # extractSim = 'examples/LmmGWAS.snps.txt',
    outFile=outFile,
    doPlots=False,
    doDebugging=True,
)



import numpy as np
import scipy.stats as st
import scipy.interpolate
import scipy.linalg as la
import fastlmm.util.preprocess as util
import scipy.stats as st
import time

# import ipdb


def print_missing_info(y):
    print(
        str(np.isnan(y).sum().sum())
        + " missing items out of "
        + str(np.product(y.shape))
        + "="
        + str(100.0 * np.isnan(y).sum().sum() / np.product(y.shape))
        + "%"
    )


# would be nice to make a low-mem version of this, but this requires
# 2-passes, because we must mean-center not only the SNPs, but also the
# individuals
def TESTBEFOREUSING_pca_covariates(bedfile, npc, outfile=None):
    """
    Read in bed file and compute PC covariates (a la Eigenstrat)
    bedfile should include the .bed extension
    returns pccov, U, S
    """
    import fastlmm.pyplink.plink as pl
    import os.path
    import time
    import scipy.linalg as la
    import numpy as np

    root, ext = os.path.splitext(bedfile)
    print("reading in bed file...")

    t0 = time.time()
    SNPs = Bed(root).read().standardize()
    t1 = time.time()
    print(("Elapsed time %.2f seconds" % (t1 - t0)))

    [N, M] = snps.shape
    print("found nind=" + str(N) + ", nsnp=" + str(M))
    if M < N:
        print("only doing full rank, should use low rank here")

    # not needed, and in practice, makes no difference
    # mean center the individuals
    # meanval=np.mean(snps,axis=0)
    # snps=snps-meanval

    print("computing kernel...")
    t0 = time.time()
    K = np.dot(snps, snps.T)
    t1 = time.time()
    print(("Elapsed time %.2f seconds" % (t1 - t0)))
    snps = None

    t0 = time.time()
    print("computing svd...")
    [U, S, V] = la.svd(K, full_matrices=False)  #!!!use big_svd?
    t1 = time.time()
    print(("Elapsed time %.2f seconds" % (t1 - t0)))

    S = np.sqrt(S)
    # UShalf=np.dot(U,np.diag(S)) #expensive
    UShalf = np.multiply(
        U, np.tile(S, (N, 1))
    )  # faster, but not as fast as as_strided which I can't get to work
    pccov = UShalf[:, 0:npc]
    print("done.")
    if outfile is not None:
        import fastlmm.util.util as ut

        ut.write_plink_covariates(SNPs["iid"], pccov, outfile)
    return pccov, U, S


def lrtpvals_qqfit_file(filein, qmax=0.1):
    import fastlmm.util.stats.chi2mixture as c2
    import fastlmm.util.stats.plotp
    import pandas as pd

    colname = "2*(LL(alt)-LL(null))"
    lrtperm = pd.read_csv(
        filein, delimiter="\t", dtype={colname: np.float64}, usecols=[colname]
    )[colname].values
    print("found " + str(len(lrtperm)) + "null test stats")

    mix = c2.chi2mixture(lrt=lrtperm, qmax=qmax, alteqnull=None)
    res = mix.fit_params_Qreg()  # parameter fitting
    print("mixture (non-zero dof)=" + str(mix.mixture) + "\n")
    print("dof=" + str(res["dof"]) + "\n")
    print("scale=" + str(res["scale"]) + "\n")


def lrt_method(stat, dof):
    """
    Standard way to compute p-value when no boundary conditions
    """
    pv = st.chi2.sf(stat, dof)
    return pv


def linreg_uniscan(X, y, covar=None, REML=False):
    """
    Iterate through each column of X, using it in the alternative model, and otherwise,
    only covar/covar+bias in the null model. Y is a 1D phenotype vector
    Output:
        p-value, LLnull, LLalt
    """
    numtests = X.shape[1]
    print("numtests=" + str(numtests))
    pv = np.nan * np.ones(numtests)
    m_alt = []
    nullfeat = np.ones_like(y)  # bias term
    if covar is not None:
        nullfeat = np.hstack((nullfeat, covar))

    m_null = linreg(nullfeat, y, REML=REML)
    ttt0 = time.time()
    for i in range(numtests):
        snp = X[:, i : i + 1]
        if np.isnan(y).sum() > 0:
            raise Exception("missing data found")
        snp_w_nullfeat = np.hstack((snp, nullfeat))
        m_alt.append(linreg(snp_w_nullfeat, y, REML=REML))
        pv[i] = lrt(-2 * (m_alt[i]["nLL"] - m_null["nLL"]), dof=1)
        if i % 50000 == 0:
            ttt1 = time.time()
            print(
                "Elapsed time for %d of %d tests is %.2f seconds"
                % (i, numtests, (ttt1 - ttt0))
            )
    return pv, m_null, m_alt


def linreg(X, y, REML=True, **kwargs):
    """
    Perform linear regression using the built-in least squares solver
    --------------------------------------------------------------------------
    Output dictionary:
    'nLL'       : negative log-likelihood
    'beta'      : [D*1] array of fixed effects weights beta
    'sigma2'        : variance of noise
    'REML'      : True: REML was computed, False: ML was computed
    --------------------------------------------------------------------------
    """

    lsqSol = la.lstsq(X, y)

    beta = lsqSol[0]  # weights
    r2 = lsqSol[1]  # squared residuals
    D = lsqSol[2]  # rank of design matrix

    N = y.shape[0]

    if not (REML):
        sigma2 = r2 / (N)
        nLL = N * 0.5 * np.log(2 * np.pi * sigma2) + N * 0.5
    else:
        sigma2 = r2 / (N - D)
        nLL = N * 0.5 * np.log(2 * np.pi * sigma2) + 0.5 / sigma2 * r2
        nLL -= 0.5 * D * np.log(2 * np.pi * sigma2)
        # REML term

    result = {"nLL": nLL, "sigma2": sigma2, "beta": beta, "REML": REML}
    if result["nLL"] is None:
        raise Exception("no nLL result")
    return result


def stats(a):
    """
    Reports the mean, std, min, and max values in a
    """
    print(
        "mean="
        + str(np.mean(a))
        + ", std="
        + str(np.std(a))
        + ", [min,max]=["
        + str(np.min(a))
        + ","
        + str(np.max(a))
        + "]"
    )



"""
simple module to save and load compressed pickle files
"""

import pickle
import bz2
import sys


def save(filename, myobj):
    """
    save object to file using pickle

    @param filename: name of destination file
    @type filename: str
    @param myobj: object to save (has to be pickleable)
    @type myobj: obj
    """

    try:
        f = bz2.BZ2File(filename, "wb")
    except IOError as details:
        sys.stderr.write("File " + filename + " cannot be written\n")
        sys.stderr.write(details)
        return

    pickle.dump(myobj, f, protocol=2)
    f.close()


def load(filename):
    """
    Load from filename using pickle

    @param filename: name of file to load from
    @type filename: str
    """

    try:
        f = bz2.BZ2File(filename, "rb")
    except IOError as details:
        sys.stderr.write("File " + filename + " cannot be read\n")
        sys.stderr.write(details)
        return

    myobj = pickle.load(f)
    f.close()
    return myobj



import numpy as np
import time
import fastlmm.util.standardizer as stdizer
import warnings


# TODO: wrap C-stuff here?
def mean_impute(X, imissX=None, maxval=2.0):
    """
    mean impute an array of floats
    ---------------------------------------------------
    X           data array
    imissX      indicator array of missing values
    maxval      larges value in X (default 2.0)
    ---------------------------------------------------
    """
    if imissX is None:
        imissX = np.isnan(X)

    n_i, n_s = X.shape
    if imissX is None:
        n_obs_SNP = np.ones(X.shape)
    else:
        i_nonan = ~imissX
        n_obs_SNP = i_nonan.sum(0)
        X[imissX] = 0.0
    snp_sum = (X).sum(0)
    one_over_sqrt_pi = (1.0 + snp_sum) / (2.0 + maxval * n_obs_SNP)
    one_over_sqrt_pi = 1.0 / np.sqrt(one_over_sqrt_pi * (1.0 - one_over_sqrt_pi))
    snp_mean = (snp_sum * 1.0) / (n_obs_SNP)

    X_ret = X - snp_mean
    X_ret *= one_over_sqrt_pi
    if imissX is not None:
        X_ret[imissX] = 0.0
    return X_ret


def standardize(
    snps, blocksize=None, standardizer=stdizer.Unit(), force_python_only=False
):
    """Does in-place standardization.
    Will use C++ if possible (for single and double, unit and beta, order="F" and order="C")
    """
    #!!warnings.warn("This standardizer is deprecated. Pysnptools includes newer versions of standardization", DeprecationWarning)
    if isinstance(standardizer, str):
        standardizer = standardizer.factor(standardizer)

    if (
        blocksize is not None and blocksize >= snps.shape[1]
    ):  # If blocksize is larger than the # of snps, set it to None
        blocksize = None

    return standardizer.standardize(
        snps, blocksize=blocksize, force_python_only=force_python_only
    )



import numpy as np
import subprocess, sys, os.path
from itertools import *
from fastlmm.pyplink.snpset import *
from fastlmm.pyplink.altset_list import *
import pandas as pd
import logging


class Dat(object):
    """
    This is a class that reads into memory from DAT/FAM/MAP files.
    """

    def __init__(self, dat_filename):
        """
        filename    : string of the name of the Dat file.
        """

        self.dat_filename = dat_filename
        self._ran_once = False

    #!! similar code in fastlmm
    def __repr__(self):
        return "{0}('{1}')".format(self.__class__.__name__, self.dat_filename)

    def names_of_other_files(self):
        return Dat.names_of_other_files_static(self.dat_filename)

    @property
    def snp_to_index(self):
        self.run_once()
        return self._snp_to_index

    @staticmethod
    def names_of_other_files_static(dat_filename):
        base_filename = (".").join(dat_filename.split(".")[:-1])
        famfile = base_filename + ".fam"
        mapfile = base_filename + ".map"
        return famfile, mapfile

    def run_once(self):
        if self._ran_once:
            return
        self._ran_once = True

        famfile, mapfile = self.names_of_other_files()

        #!!similar code in BED reader
        logging.info("Loading fam file {0}".format(famfile))
        self._original_iids = np.loadtxt(
            famfile, dtype="str", usecols=(0, 1), comments=None
        )

        #!!similar code in BED reader
        logging.info("Loading map file {0}".format(mapfile))
        self.bimfields = pd.read_csv(
            mapfile,
            delimiter="\s",
            usecols=(0, 1, 2, 3),
            header=None,
            index_col=False,
            engine="python",
        )
        self.rs = np.array(self.bimfields[1].tolist(), dtype="str")
        self.pos = self.bimfields[[0, 2, 3]].values
        self._snp_to_index = {}
        logging.info("indexing snps")
        for i in range(self.snp_count):
            snp = self.rs[i]
            if snp in self._snp_to_index:
                raise Exception(
                    "Expect snp to appear in bim file only once. ({0})".format(snp)
                )
            self._snp_to_index[snp] = i

        #!!could change to just create/find an index to the file position of each row. Instead, reading all into memory
        datfields = pd.read_csv(
            self.dat_filename, delimiter="\s", header=None, index_col=False
        )
        if not np.array_equal(np.array(datfields[0], dtype="string"), self.rs):
            raise Exception(
                "Expect snp list in map file to exactly match snp list in dat file"
            )
        self.start_column = 3
        if len(self._original_iids) != datfields.shape[1] - self.start_column:
            raise Exception("Expect # iids in fam file to match dat file")
        self.datfields = datfields.T

        return self

    # def __del__(self):
    #    if self._filepointer is not None:  # we need to test this because Python doesn't guarantee that __init__ was fully run
    #        self._filepointer.close()

    def copyinputs(self, copier):
        # doesn't need to self.run_once() because creates name of all files itself
        famfile, mapfile = self.names_of_other_files()
        copier.input(self.dat_filename)
        copier.input(famfile)
        copier.input(mapfile)

    @property
    def snp_count(self):
        self.run_once()
        return len(self.bimfields)

    def read(
        self, snp_set=AllSnps(), order="F", dtype=np.float64, force_python_only=False
    ):
        """
        Input: a snp_set. Choices include
            AllSnps() [the default],
            PositionRange(snpIndex,nSNPs)
            SnpAndSetName(groupname,snplist),

        Output dictionary:
        'rs'     : [S] array rs-numbers
        'pos'    : [S*3] array of positions [chromosome, genetic dist, basepair dist]
        'snps'   : [N*S] array of snp-data
        'iid'    : [N*2] array of family IDs and individual IDs
        """
        self.run_once()
        snpset_withbed = snp_set.addbed(self)
        return self.read_with_specification(
            snpset_withbed,
            order=order,
            dtype=dtype,
            force_python_only=force_python_only,
        )

    ##!! This property is ugly
    @property
    def ind_used(self):
        # doesn't need to self.run_once() because only uses original inputs
        return self._ind_used

    @ind_used.setter
    def ind_used(self, value):
        """
        Tell the Dat reader to return data for only a subset (perhaps proper) of the individuals in a particular order
        e.g. 2,10,0 says to return data for three users: the user at index position 2, the user at index position 10, and the user at index position 0.
        """
        # doesn't need to self.run_once() because only uses original inputs
        self._ind_used = value

    @property
    def original_iids(self):
        self.run_once()
        return self._original_iids

    def counts_and_indexes(self, snpset_withdat):
        #!!similar code in BED
        iid_count_in = len(self.original_iids)
        snp_count_in = self.snp_count
        if hasattr(self, "_ind_used"):
            iid_count_out = len(self.ind_used)
            iid_index_out = self.ind_used
        else:
            iid_count_out = iid_count_in
            iid_index_out = list(range(0, iid_count_in))
        snp_count_out = len(snpset_withdat)
        snp_index_out = list(
            snpset_withdat
        )  # make a copy, in case it's in some strange format, such as HDF5
        return (
            iid_count_in,
            iid_count_out,
            iid_index_out,
            snp_count_in,
            snp_count_out,
            snp_index_out,
        )

    @staticmethod
    def read_with_specification(
        snpset_withdat, order="F", dtype=np.float64, force_python_only=False
    ):
        dat = snpset_withdat.bed  #!!bed is a misnomer
        (
            iid_count_in,
            iid_count_out,
            iid_index_out,
            snp_count_in,
            snp_count_out,
            snp_index_out,
        ) = dat.counts_and_indexes(snpset_withdat)

        SNPs = np.zeros((iid_count_out, snp_count_out), order=order, dtype=dtype)
        for SNPsIndex, bimIndex in enumerate(snp_index_out):
            row = np.array(dat.datfields[bimIndex])[dat.start_column :,]
            SNPs[:, SNPsIndex] = row[iid_index_out]

        ret = {
            "rs": dat.rs[snp_index_out],
            "pos": dat.pos[snp_index_out, :],
            "snps": SNPs,
            "iid": dat.original_iids[iid_index_out, :],
        }
        return ret

    @staticmethod
    def write(snpMatrix, datfile):
        snpsarray = snpMatrix["snps"]
        rslist = snpMatrix["rs"]
        posarray = snpMatrix["pos"]
        iidarray = snpMatrix["iid"]

        famfile, mapfile = Dat.names_of_other_files_static(datfile)

        with open(famfile, "w") as fam_filepointer:
            for iid_row in iidarray:
                fam_filepointer.write(
                    "{0} {1} 0 0 0 0\n".format(iid_row[0], iid_row[1])
                )

        with open(mapfile, "w") as map_filepointer:
            for snp_index, rs in enumerate(rslist):
                posrow = posarray[snp_index]
                map_filepointer.write(
                    "{0}\t{1}\t{2}\t{3}\n".format(posrow[0], rs, posrow[1], posrow[2])
                )

        with open(datfile, "w") as dat:
            for snp_index, rs in enumerate(rslist):
                dat.write(
                    "{0}\tj\tn\t".format(rs)
                )  # use "j" and "n" as the major and minor allele
                row = snpsarray[:, snp_index]
                dat.write("\t".join([str(i) for i in row]) + "\n")


# if __name__ == "__main__":
#    logging.basicConfig(level=logging.INFO)
#    import doctest
#    doctest.testmod()



import fastlmm.association.lrt as lr
import numpy as np
import fastlmm.util.stats.chi2mixture as c2

from . import tests_util as tu


class Lrt(object):
    """description of class"""

    def check_nperm(self, nperm):
        return nperm  # permutations are fine, so just return

    def __str__(self):
        return "lrt"

    def construct(
        self,
        Y,
        X=None,
        forcefullrank=False,
        SNPs0=None,
        i_exclude=None,
        nullModel=None,
        altModel=None,
        scoring=None,
        greater_is_better=None,
    ):
        G0, K0 = tu.set_snps0(SNPs0=SNPs0, sample_size=Y.shape[0], i_exclude=i_exclude)
        return lr.lrt(
            Y=Y,
            X=X,
            model0=None,
            appendbias=False,
            forcefullrank=forcefullrank,
            G0=G0,
            K0=K0,
            nullModel=nullModel,
            altModel=altModel,
        )

    def construct_no_backgound_kernel(
        self, Y, X, forcefullrank, nullModel, altModel, scoring, greater_is_better
    ):
        return self.construct(
            Y=Y,
            X=X,
            forcefullrank=forcefullrank,
            nullModel=nullModel,
            altModel=altModel,
            scoring=scoring,
            greater_is_better=greater_is_better,
        )

    def pv(squaredform, expectationsqform, varsqform, GPG):
        raise Exception("'pv' doesn't apply to lrt only to davies")

    @property
    def npvals(self):
        return 1  # return only 1 type of p-value

    def w2(self, G0, result):
        if G0 is not None:
            return result.a2
        else:
            return result.h2

    def lrt_method(self, result):
        return result.stat

    def pv_adj_from_result(self, result):
        """
        If local aUD exists, take that, if not, take the raw local.
        """
        if "pv-local-aUD" in result.test and not np.isnan(result.test["pv-local-aUD"]):
            return result.test["pv-local-aUD"]
        elif "pv-local" in result.test:
            return result.test["pv-local"]
        else:
            return np.nan

    def pv_adj_and_ind(
        self,
        nperm,
        pv_adj,
        nullfit,
        lrt,
        lrtperm,
        alteqnull,
        alteqnullperm,
        qmax,
        nullfitfile,
        nlocalperm,
    ):
        if nlocalperm and nlocalperm > 0:  # don't do the fitting
            ind = pv_adj.argsort()
            return pv_adj, ind

        from fastlmm.association.tests import Cv

        return Cv.pv_adj_and_ind(
            nperm,
            pv_adj,
            nullfit,
            lrt,
            lrtperm,
            alteqnull,
            alteqnullperm,
            qmax,
            nullfitfile,
            nlocalperm,
        )  # call the shared version of this method

    def write(self, fp, ind, result_dict, pv_adj, detailed_table, signal_ratio=True):
        if "pv-local-aUD" in result_dict[0].test:
            # in this case, for p_adj, we use pv-local-aUD if it exists, and otherwise
            # pv-local. So don't know which is which in the "P-value adjusted" column. To
            # disambiguate, also print out "pv-local" here
            colnames = [
                "SetId",
                "LogLikeAlt",
                "LogLikeNull",
                "P-value_adjusted",
                "P-value-local",
                "P-value(50/50)",
                "#SNPs_in_Set",
                "#ExcludedSNPs",
                "chrm",
                "pos. range",
            ]
        else:
            colnames = [
                "SetId",
                "LogLikeAlt",
                "LogLikeNull",
                "P-value_adjusted",
                "P-value(50/50)",
                "#SNPs_in_Set",
                "#ExcludedSNPs",
                "chrm",
                "pos. range",
            ]
        if signal_ratio:
            colnames.append("Alt_h2")
            colnames.append("Alt_a2")

        head = "\t".join(colnames)

        if detailed_table:
            lik1Info = result_dict[0].lik1Details
            lik0Info = result_dict[0].lik0Details

            altNames = list(lik1Info.keys())
            altIndices = sorted(list(range(len(altNames))), key=lambda k: altNames[k])
            altNames.sort()

            altNames = ["Alt" + t for t in altNames]
            head += "\t" + "\t".join(altNames)

            nullNames = list(lik0Info.keys())
            nullIndices = sorted(
                list(range(len(nullNames))), key=lambda k: nullNames[k]
            )
            nullNames.sort()

            nullNames = ["Null" + t for t in nullNames]
            head += "\t" + "\t".join(nullNames)

        head += "\n"

        fp.write(head)

        for i in range(len(ind)):
            ii = ind[i]
            result = result_dict[ii]
            ll0 = str(-(result.stat / 2.0 + result.test["lik1"]["nLL"]))

            if "pv-local-aUD" in result_dict[0].test:
                rowvals = [
                    result.setname,
                    str(-result.test["lik1"]["nLL"]),
                    ll0,
                    str(pv_adj[ii]),
                    str(result.test["pv-local"]),
                    str(result.pv),
                    str(result.setsize),
                    str(result.nexclude),
                    result.ichrm,
                    result.iposrange,
                ]
            else:
                rowvals = [
                    result.setname,
                    str(-result.test["lik1"]["nLL"]),
                    ll0,
                    str(pv_adj[ii]),
                    str(result.pv),
                    str(result.setsize),
                    str(result.nexclude),
                    result.ichrm,
                    result.iposrange,
                ]

            if signal_ratio:
                rowvals.append(str(result.h2))
                rowvals.append(str(result.a2))

            row = "\t".join(rowvals)

            if detailed_table:
                lik1Info = result.lik1Details
                lik0Info = result.lik0Details

                vals = list(lik1Info.values())
                vals = [vals[j] for j in altIndices]
                row += "\t" + "\t".join([str(v) for v in vals])

                vals = list(lik0Info.values())
                vals = [vals[j] for j in nullIndices]
                row += "\t" + "\t".join([str(v) for v in vals])

            row += "\n"
            fp.write(row)

    def pv_etc(
        self, filenull, G0_to_use, G1, y, x, null_model, varcomp_test, forcefullrank
    ):
        if self.filenull is not None:
            return lr.twokerneltest(
                G0=G0_to_use,
                G1=G1,
                y=y,
                covar=x,
                appendbias=False,
                lik0=null_model,
                forcefullrank=forcefullrank,
            )
        else:
            return lr.onekerneltest(
                G1=G1,
                y=y,
                covar=x,
                appendbias=False,
                lik0=varcomp_test,
                forcefullrank=self.forcefullrank,
            )
